@@ -26,7 +26,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: defs.h,v 1.9 1999/12/23 15:08:17 wichert Exp $
+ *	$Id: defs.h,v 1.24 2001/03/28 08:05:27 hughesj Exp $
  */
 
 #ifdef linux
@@ -39,10 +39,14 @@
 
 /* configuration section */
 #ifndef MAX_QUALS
-#define MAX_QUALS	1024	/* maximum number of syscalls, signals, etc. */
+#if defined(linux) && defined(MIPS)
+#define MAX_QUALS	4999	/* maximum number of syscalls, signals, etc. */
+#else
+#define MAX_QUALS	2048	/* maximum number of syscalls, signals, etc. */
+#endif
 #endif
 #ifndef MAX_PROCS
-#define MAX_PROCS	32	/* maximum number of processes tracable */
+#define MAX_PROCS	64	/* maximum number of processes tracable */
 #endif
 #ifndef DEFAULT_STRLEN
 #define DEFAULT_STRLEN	32	/* default maximum # of bytes printed in
@@ -64,6 +68,7 @@
 #include <stdio.h>
 #include <ctype.h>
 #include <string.h>
+#include <time.h>
 #include <sys/time.h>
 #include <errno.h>
 
@@ -80,12 +85,30 @@
 #  endif
 #endif 
 
-#ifdef SVR4
+#if defined(SVR4) || defined(FREEBSD)
+#define USE_PROCFS
+#else
+#undef USE_PROCFS
+#endif
+
+#ifdef FREEBSD
+#ifndef I386
+#error "FreeBSD support is only for i386 arch right now."
+#endif
+#include <machine/psl.h>
+#include <machine/reg.h>
+#include <sys/syscall.h>
+#endif
+
+#ifdef USE_PROCFS
 #include <sys/procfs.h>
 #ifdef HAVE_MP_PROCFS
 #include <sys/uio.h>
 #endif
-#else /* !SVR4 */
+#ifdef FREEBSD
+#include <sys/pioctl.h>
+#endif /* FREEBSD */
+#else /* !USE_PROCFS */
 #if defined(LINUXSPARC) && defined(__GLIBC__)
 #include <sys/ptrace.h>
 #else
@@ -134,6 +157,13 @@ extern int ptrace();
 #  define REG_SP 29
 #  define REG_EPC 64
 #endif /* MIPS */
+#ifdef HPPA
+#  define PT_GR20 (20*4)
+#  define PT_GR26 (26*4)
+#  define PT_GR28 (28*4)
+#  define PT_IAOQ0 (106*4)
+#  define PT_IAOQ1 (107*4)
+#endif /* HPPA */
 #endif /* LINUX */
 
 #define SUPPORTED_PERSONALITIES 1
@@ -154,12 +184,13 @@ extern int mp_ioctl (int f, int c, void *a, int s);
 #define IOCTL_STATUS(t) \
 	 pread (t->pfd_stat, &t->status, sizeof t->status, 0)
 #define IOCTL_WSTOP(t)						\
-	(IOCTL (t->pfd, PCWSTOP, NULL) < 0 ? -1 :		\
+	(IOCTL (t->pfd, PCWSTOP, (char *)NULL) < 0 ? -1 :		\
 	 IOCTL_STATUS (t))
 #define PR_WHY		pr_lwp.pr_why
 #define PR_WHAT		pr_lwp.pr_what
 #define PR_REG		pr_lwp.pr_context.uc_mcontext.gregs
 #define PR_FLAGS	pr_lwp.pr_flags
+#define PR_SYSCALL	pr_lwp.pr_syscall
 #define PIOCSTIP	PCSTOP
 #define PIOCSET		PCSET
 #define PIOCRESET	PCRESET
@@ -178,8 +209,26 @@ extern int mp_ioctl (int f, int c, void *a, int s);
 #define PR_WHAT		pr_what
 #define PR_REG		pr_reg
 #define PR_FLAGS	pr_flags
+#define PR_SYSCALL	pr_syscall
 #endif
 #endif
+#ifdef FREEBSD
+#define IOCTL		ioctl
+#define IOCTL_STATUS(t)	ioctl (t->pfd, PIOCSTATUS, &t->status)
+#define IOCTL_WSTOP(t)	ioctl (t->pfd, PIOCWAIT, &t->status)
+#define PIOCRUN         PIOCCONT
+#define PIOCWSTOP       PIOCWAIT
+#define PR_WHY		why
+#define PR_WHAT		val
+#define PR_FLAGS	state
+/* from /usr/src/sys/miscfs/procfs/procfs_vnops.c,
+   status.state = 0 for running, 1 for stopped */
+#define PR_ASLEEP	1
+#define PR_SYSENTRY     S_SCE
+#define PR_SYSEXIT      S_SCX
+#define PR_SIGNALLED    S_SIG
+#define PR_FAULTED      S_CORE
+#endif
 
 /* Trace Control Block */
 struct tcb {
@@ -190,8 +239,11 @@ struct tcb {
 	long u_arg[MAX_ARGS];	/* System call arguments */
 	int u_error;		/* Error code */
 	long u_rval;		/* (first) return value */
+#ifdef HAVE_LONG_LONG
+	long long u_lrval;	/* long long return value */
+#endif
 	FILE *outf;		/* Output file for this process */
-	char *auxstr;		/* Auxiliary info from syscall (see RVAL_STR) */
+	const char *auxstr;	/* Auxiliary info from syscall (see RVAL_STR) */
 	struct timeval stime;	/* System time usage as of last process wait */
 	struct timeval dtime;	/* Delta for system time usage */
 	struct timeval etime;	/* Syscall entry time */
@@ -212,6 +264,11 @@ struct tcb {
 	prstatus_t status;	/* procfs status structure */
 #endif
 #endif
+#ifdef FREEBSD
+	struct procfs_status status;
+	int pfd_reg;
+	int pfd_status;
+#endif
 };
 
 /* TCB flags */
@@ -227,7 +284,7 @@ struct tcb {
 #define TCB_FOLLOWFORK	00400	/* Process should have forks followed */
 #define TCB_REPRINT	01000	/* We should reprint this syscall on exit */
 #ifdef LINUX
-#if defined(ALPHA) || defined(SPARC) || defined(POWERPC)
+#if defined(ALPHA) || defined(SPARC) || defined(POWERPC) || defined(HPPA)
 #define TCB_WAITEXECVE	02000	/* ignore SIGTRAP after exceve */
 #endif /* ALPHA */
 #endif /* LINUX */
@@ -262,6 +319,10 @@ struct xlat {
 #define RVAL_HEX	001	/* hex format */
 #define RVAL_OCTAL	002	/* octal format */
 #define RVAL_UDECIMAL	003	/* unsigned decimal format */
+#define RVAL_LDECIMAL	004	/* long decimal format */
+#define RVAL_LHEX	005	/* long hex format */
+#define RVAL_LOCTAL	006	/* long octal format */
+#define RVAL_LUDECIMAL	007	/* long unsigned decimal format */
 #define RVAL_MASK	007	/* mask for these values */
 
 #define RVAL_STR	010	/* Print `auxstr' field after return val */
@@ -306,6 +367,7 @@ extern void set_sortby P((char *));
 extern void set_overhead P((int));
 extern void qualify P((char *));
 extern void newoutf P((struct tcb *));
+extern int get_scno P((struct tcb *));
 extern int trace_syscall P((struct tcb *));
 extern void printxval P((struct xlat *, int, char *));
 extern int printargs P((struct tcb *));
@@ -366,7 +428,7 @@ extern int fixvfork P((struct tcb *));
 #if !(defined(LINUX) && !defined(SPARC))
 extern long getrval2 P((struct tcb *));
 #endif
-#ifdef SVR4
+#ifdef USE_PROCFS
 extern int proc_open P((struct tcb *tcp, int attaching));
 #endif
 
@@ -441,3 +503,30 @@ extern int nioctlents2;
 extern char *signalent2[];
 extern int nsignals2;
 #endif /* SUPPORTED_PERSONALITIES >= 3 */
+
+#if FREEBSD
+/* ARRGH!  off_t args are aligned on 64 bit boundaries! */
+#define ALIGN64(tcp,arg)						\
+do {									\
+	if (arg % 2)							\
+	    memmove (&tcp->u_arg[arg], &tcp->u_arg[arg + 1],		\
+		     (tcp->u_nargs - arg - 1) * sizeof tcp->u_arg[0]);	\
+} while (0)
+#else
+#define ALIGN64(tcp,arg) do { } while (0)
+#endif
+
+#if HAVE_LONG_LONG
+
+/* _l refers to the lower numbered u_arg,
+ * _h refers to the higher numbered u_arg
+ */
+
+#if HAVE_LITTLE_ENDIAN_LONG_LONG
+#define LONG_LONG(_l,_h) \
+    ((long long)((unsigned long long)(_l) | ((unsigned long long)(_h)<<32)))
+#else
+#define LONG_LONG(_l,_h) \
+    ((long long)((unsigned long long)(_h) | ((unsigned long long)(_l)<<32)))
+#endif
+#endif
