@@ -25,7 +25,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: syscall.h,v 1.9 1999/12/24 07:59:50 drepper Exp $
+ *	$Id: syscall.h,v 1.13 2001/03/27 12:17:17 wichert Exp $
  */
 
 #include "dummy.h"
@@ -71,7 +71,7 @@ int sys_setfsuid(), sys_setfsgid(), sys_llseek();
 int sys_getdents(), sys_flock(), sys_msync();
 int sys_readv(), sys_writev(), sys_select();
 int sys_getsid(), sys_fdatasync(), sys_sysctl();
-int sys_mlock(), sys_munlock(), sys_mlockall(), sys_munlockall();
+int sys_mlock(), sys_munlock(), sys_mlockall(), sys_munlockall(), sys_madvise();
 int sys_sched_setparam(), sys_sched_getparam();
 int sys_sched_setscheduler(), sys_sched_getscheduler(), sys_sched_yield();
 int sys_sched_get_priority_max(), sys_sched_get_priority_min();
@@ -82,7 +82,7 @@ int sys_sigaltstack(), sys_rt_sigprocmask(), sys_rt_sigaction();
 int sys_rt_sigpending(), sys_rt_sigsuspend(), sys_rt_sigqueueinfo();
 int sys_rt_sigtimedwait(), sys_prctl(), sys_poll(), sys_vfork();
 int sys_sendfile(), sys_old_mmap(), sys_stat64(), sys_lstat64(), sys_fstat64();
-int sys_truncate64(), sys_ftruncate64();
+int sys_truncate64(), sys_ftruncate64(), sys_pivotroot();
 
 
 /* sys_socketcall subcalls */
@@ -109,7 +109,7 @@ int sys_osf_utimes();
 #endif
 
 
-#if !defined(ALPHA) && !defined(MIPS)
+#if !defined(ALPHA) && !defined(IA64) && !defined(MIPS) &&!defined(HPPA)
 #ifdef POWERPC
 #  define SYS_socket_subcall	256
 #else
@@ -134,7 +134,7 @@ int sys_osf_utimes();
 #define SYS_recvmsg		(SYS_socket_subcall + 17)
 
 #define SYS_socket_nsubcalls	18
-#endif /* !ALPHA */
+#endif /* !(ALPHA || IA64 || MIPS) */
 
 /* sys_ipc subcalls */
 
@@ -142,11 +142,11 @@ int sys_semget(), sys_semctl(), sys_semop();
 int sys_msgsnd(), sys_msgrcv(), sys_msgget(), sys_msgctl();
 int sys_shmat(), sys_shmdt(), sys_shmget(), sys_shmctl();
 
-#if !defined(ALPHA) && !defined(MIPS)
+#if !defined(ALPHA) && !defined(IA64) && !defined(MIPS) && !defined(SPARC) &&!defined(HPPA)
 #ifdef POWERPC
 #  define SYS_ipc_subcall		((SYS_socket_subcall)+(SYS_socket_nsubcalls))
 #else
-#  define SYS_ipc_subcall		200
+#  define SYS_ipc_subcall		250
 #endif
 #define SYS_semop		(SYS_ipc_subcall + 1)
 #define SYS_semget		(SYS_ipc_subcall + 2)
@@ -161,10 +161,22 @@ int sys_shmat(), sys_shmdt(), sys_shmget(), sys_shmctl();
 #define SYS_shmctl		(SYS_ipc_subcall + 24)
 
 #define SYS_ipc_nsubcalls	25
-#endif /* ALPHA */
+#endif /* !(ALPHA || IA64 || MIPS) */
+
+#if defined(ALPHA) || defined(IA64)
+int sys_getpagesize();
+#endif
 
 #ifdef ALPHA
-int osf_statfs(), osf_fstatfs(), sys_getpagesize(), sys_madvise();
+int osf_statfs(), osf_fstatfs();
+#endif
+
+#ifdef IA64
+int sys_getpmsg(), sys_putpmsg();	/* STREAMS stuff */
+#endif
+
+#ifdef MIPS
+int sys_sysmips()
 #endif
 
 int sys_setpgrp(), sys_gethostname(), sys_getdtablesize(), sys_utimes();
