@@ -2,7 +2,7 @@
  * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
  * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
  * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
- * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
+ * Copyright (c) 1996-2000 Wichert Akkerman <wichert@cistron.nl>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -27,7 +27,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: net.c,v 1.12 1999/12/23 14:20:15 wichert Exp $
+ *	$Id: net.c,v 1.23 2001/03/27 16:47:36 hughesj Exp $
  */
 
 #include "defs.h"
@@ -35,8 +35,22 @@
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/un.h>
+
+#if defined(HAVE_SIN6_SCOPE_ID_LINUX)
+#define in6_addr in6_addr_libc
+#define ipv6_mreq ipv6_mreq_libc
+#define sockaddr_in6 sockaddr_in6_libc
+#endif
+
 #include <netinet/in.h>
+#ifdef HAVE_NETINET_TCP_H
+#include <netinet/tcp.h>
+#endif
+#ifdef HAVE_NETINET_UDP_H
+#include <netinet/udp.h>
+#endif
 #include <arpa/inet.h>
+#include <net/if.h>
 #if defined(LINUX)
 #include <asm/types.h>
 #if defined(__GLIBC__) && (__GLIBC__ >= 2) && (__GLIBC__ + __GLIBC_MINOR__ >= 3)
@@ -46,9 +60,25 @@
 #endif
 #endif /* LINUX */
 
-#if defined (__GLIBC__) && ((__GLIBC__ < 2) || (__GLIBC__ == 2 && __GLIBC_MINOR__ < 1))
+#if defined (__GLIBC__) && (((__GLIBC__ < 2) || (__GLIBC__ == 2 && __GLIBC_MINOR__ < 1)) || defined(HAVE_SIN6_SCOPE_ID_LINUX))
 #if defined(HAVE_LINUX_IN6_H)
+#if defined(HAVE_SIN6_SCOPE_ID_LINUX)
+#undef in6_addr
+#undef ipv6_mreq
+#undef sockaddr_in6
+#define in6_addr in6_addr_kernel
+#define ipv6_mreq ipv6_mreq_kernel
+#define sockaddr_in6 sockaddr_in6_kernel
+#endif
 #include <linux/in6.h>
+#if defined(HAVE_SIN6_SCOPE_ID_LINUX)
+#undef in6_addr
+#undef ipv6_mreq
+#undef sockaddr_in6
+#define in6_addr in6_addr_libc
+#define ipv6_mreq ipv6_mreq_libc
+#define sockaddr_in6 sockaddr_in6_kernel
+#endif
 #endif
 #endif
 
@@ -64,10 +94,18 @@
 #include <linux/if_packet.h>
 #endif
 
+#if defined(HAVE_LINUX_ICMP_H)
+#include <linux/icmp.h>
+#endif
+
 #ifndef PF_UNSPEC
 #define PF_UNSPEC AF_UNSPEC
 #endif
 
+#if UNIXWARE >= 7
+#define HAVE_SENDMSG		1		/* HACK - *FIXME* */
+#endif
+
 #ifdef LINUX
 /* Under Linux these are enums so we can't test for them with ifdef. */
 #define IPPROTO_EGP IPPROTO_EGP
@@ -180,6 +218,66 @@ static struct xlat socktypes[] = {
 #endif
 	{ 0,		NULL		},
 };
+static struct xlat socketlayers[] = {
+#if defined(SOL_IP)
+	{ SOL_IP,	"SOL_IP"	},
+#endif
+#if defined(SOL_ICMP)
+	{ SOL_ICMP,	"SOL_ICMP"	},
+#endif
+#if defined(SOL_TCP)
+	{ SOL_TCP,	"SOL_TCP"	},
+#endif
+#if defined(SOL_UDP)
+	{ SOL_UDP,	"SOL_UDP"	},
+#endif
+#if defined(SOL_IPV6)
+	{ SOL_IPV6,	"SOL_IPV6"	},
+#endif
+#if defined(SOL_ICMPV6)
+	{ SOL_ICMPV6,	"SOL_ICMPV6"	},
+#endif
+#if defined(SOL_RAW)
+	{ SOL_RAW,	"SOL_RAW"	},
+#endif
+#if defined(SOL_IPX)
+	{ SOL_IPX,	"SOL_IPX"	},
+#endif
+#if defined(SOL_IPX)
+	{ SOL_IPX,	"SOL_IPX"	},
+#endif
+#if defined(SOL_AX25)
+	{ SOL_AX25,	"SOL_AX25"	},
+#endif
+#if defined(SOL_ATALK)
+	{ SOL_ATALK,	"SOL_ATALK"	},
+#endif
+#if defined(SOL_NETROM)
+	{ SOL_NETROM,	"SOL_NETROM"	},
+#endif
+#if defined(SOL_ROSE)
+	{ SOL_ROSE,	"SOL_ROSE"	},
+#endif
+#if defined(SOL_DECNET)
+	{ SOL_DECNET,	"SOL_DECNET"	},
+#endif
+#if defined(SOL_X25)
+	{ SOL_X25,	"SOL_X25"	},
+#endif
+#if defined(SOL_PACKET)
+	{ SOL_PACKET,	"SOL_PACKET"	},
+#endif
+#if defined(SOL_ATM)
+	{ SOL_ATM,	"SOL_ATM"	},
+#endif
+#if defined(SOL_AAL)
+	{ SOL_AAL,	"SOL_AAL"	},
+#endif
+#if defined(SOL_IRDA)
+	{ SOL_IRDA,	"SOL_IRDA"	},
+#endif
+	{ SOL_SOCKET,	"SOL_SOCKET"	},	/* Never used! */
+};
 static struct xlat protocols[] = {
 	{ IPPROTO_IP,	"IPPROTO_IP"	},
 	{ IPPROTO_ICMP,	"IPPROTO_ICMP"	},
@@ -351,20 +449,35 @@ static struct xlat sockoptions[] = {
 
 #ifdef SOL_IP
 static struct xlat sockipoptions[] = {
-	{ IP_TOS,       "IP_TOS"        },
-	{ IP_TTL,       "IP_TTL"        },
+	{ IP_TOS,		"IP_TOS"		},
+	{ IP_TTL,		"IP_TTL"		},
 #if defined(IP_HDRINCL)
-	{ IP_HDRINCL,   "IP_HDRINCL"    },
+	{ IP_HDRINCL,		"IP_HDRINCL"		},
 #endif
 #if defined(IP_OPTIONS)
-	{ IP_OPTIONS,   "IP_OPTIONS"    },
+	{ IP_OPTIONS,		"IP_OPTIONS"		},
 #endif
-	{ IP_MULTICAST_IF,      "IP_MULTICAST_IF"       },
-	{ IP_MULTICAST_TTL,     "IP_MULTICAST_TTL"      },
-	{ IP_MULTICAST_LOOP,    "IP_MULTICAST_LOOP"     },
-	{ IP_ADD_MEMBERSHIP,    "IP_ADD_MEMBERSHIP"     },
-	{ IP_DROP_MEMBERSHIP,   "IP_DROP_MEMBERSHIP"    },
-	{ 0,            NULL            },
+	{ IP_ROUTER_ALERT,	"IP_ROUTER_ALERT"	},
+#if defined(IP_RECVOPTIONS)
+	{ IP_RECVOPTIONS,	"IP_RECVOPTIONS"	},
+#endif
+	{ IP_RETOPTS,		"IP_RETOPTS"		},
+	{ IP_PKTINFO,		"IP_PKTINFO"		},
+	{ IP_PKTOPTIONS,	"IP_PKTOPTIONS"	},
+	{ IP_MTU_DISCOVER,	"IP_MTU_DISCOVER"	},
+	{ IP_MTU_DISCOVER,	"IP_MTU_DISCOVER"	},
+	{ IP_RECVERR,		"IP_RECVERR"		},
+	{ IP_RECVTTL,		"IP_RECRECVTTL"		},
+	{ IP_RECVTOS,		"IP_RECRECVTOS"		},
+#if defined(IP_MTU)
+	{ IP_MTU,		"IP_MTU"		},
+#endif
+	{ IP_MULTICAST_IF,	"IP_MULTICAST_IF"	},
+	{ IP_MULTICAST_TTL,	"IP_MULTICAST_TTL"	},
+	{ IP_MULTICAST_LOOP,	"IP_MULTICAST_LOOP"	},
+	{ IP_ADD_MEMBERSHIP,	"IP_ADD_MEMBERSHIP"	},
+	{ IP_DROP_MEMBERSHIP,	"IP_DROP_MEMBERSHIP"	},
+	{ 0,			NULL			},
 };
 #endif /* SOL_IP */
 
@@ -375,14 +488,89 @@ static struct xlat sockipxoptions[] = {
 };
 #endif /* SOL_IPX */
 
+#ifdef SOL_RAW
+static struct xlat sockrawoptions[] = {
+#if defined(ICMP_FILTER)
+	{ ICMP_FILTER,		"ICMP_FILTER"	},
+#endif
+	{ 0,			NULL		},
+};
+#endif /* SOL_RAW */
+
+#ifdef SOL_PACKET
+static struct xlat sockpacketoptions[] = {
+	{ PACKET_ADD_MEMBERSHIP,	"PACKET_ADD_MEMBERSHIP"	},
+	{ PACKET_DROP_MEMBERSHIP,	"PACKET_DROP_MEMBERSHIP"},
+#if defined(PACKET_RECV_OUTPUT)
+	{ PACKET_RECV_OUTPUT,		"PACKET_RECV_OUTPUT"	},
+#endif
+#if defined(PACKET_RX_RING)
+	{ PACKET_RX_RING,		"PACKET_RX_RING"	},
+#endif
+#if defined(PACKET_STATISTICS)
+	{ PACKET_STATISTICS,		"PACKET_STATISTICS"	},
+#endif
+	{ 0,				NULL			},
+};
+#endif /* SOL_PACKET */
+
 #ifdef SOL_TCP
 static struct xlat socktcpoptions[] = {
-	{ TCP_NODELAY,  "TCP_NODELAY"   },
-	{ TCP_MAXSEG,   "TCP_MAXSEG"    },
-	{ 0,            NULL            },
+	{ TCP_NODELAY,	"TCP_NODELAY"	},
+	{ TCP_MAXSEG,	"TCP_MAXSEG"	},
+#if defined(TCP_CORK)
+	{ TCP_CORK,	"TCP_CORK"	},
+#endif
+	{ 0,		NULL		},
 };
 #endif /* SOL_TCP */
 
+#ifdef SOL_RAW
+static struct xlat icmpfilterflags[] = {
+#if defined(ICMP_ECHOREPLY)
+	{ (1<<ICMP_ECHOREPLY),		"ICMP_ECHOREPLY"	},
+#endif
+#if defined(ICMP_DEST_UNREACH)
+	{ (1<<ICMP_DEST_UNREACH),	"ICMP_DEST_UNREACH"	},
+#endif
+#if defined(ICMP_SOURCE_QUENCH)
+	{ (1<<ICMP_SOURCE_QUENCH),	"ICMP_SOURCE_QUENCH"	},
+#endif
+#if defined(ICMP_REDIRECT)
+	{ (1<<ICMP_REDIRECT),		"ICMP_REDIRECT"		},
+#endif
+#if defined(ICMP_ECHO)
+	{ (1<<ICMP_ECHO),		"ICMP_ECHO"		},
+#endif
+#if defined(ICMP_TIME_EXCEEDED)
+	{ (1<<ICMP_TIME_EXCEEDED),	"ICMP_TIME_EXCEEDED"	},
+#endif
+#if defined(ICMP_PARAMETERPROB)
+	{ (1<<ICMP_PARAMETERPROB),	"ICMP_PARAMETERPROB"	},
+#endif
+#if defined(ICMP_TIMESTAMP)
+	{ (1<<ICMP_TIMESTAMP),		"ICMP_TIMESTAMP"	},
+#endif
+#if defined(ICMP_TIMESTAMPREPLY)
+	{ (1<<ICMP_TIMESTAMPREPLY),	"ICMP_TIMESTAMPREPLY"	},
+#endif
+#if defined(ICMP_INFO_REQUEST)
+	{ (1<<ICMP_INFO_REQUEST),	"ICMP_INFO_REQUEST"	},
+#endif
+#if defined(ICMP_INFO_REPLY)
+	{ (1<<ICMP_INFO_REPLY),		"ICMP_INFO_REPLY"	},
+#endif
+#if defined(ICMP_ADDRESS)
+	{ (1<<ICMP_ADDRESS),		"ICMP_ADDRESS"		},
+#endif
+#if defined(ICMP_ADDRESSREPLY)
+	{ (1<<ICMP_ADDRESSREPLY),	"ICMP_ADDRESSREPLY"	},
+#endif
+	{ 0,				NULL			},
+};
+#endif /* SOL_RAW */
+
+
 void
 printsock(tcp, addr, addrlen)
 struct tcb *tcp;
@@ -434,9 +622,9 @@ int addrlen;
 		if (addrlen==2) {
 			tprintf("<nil>");
 		} else if (addrbuf.sau.sun_path[0]) {
-			tprintf("path=\"%*.*s\"", addrlen-2, addrlen-2, addrbuf.sau.sun_path);
+			tprintf("path=\"%-.*s\"", addrlen-2, addrbuf.sau.sun_path);
 		} else {
-			tprintf("path=@%*.*s", addrlen-3, addrlen-3, addrbuf.sau.sun_path+1);
+			tprintf("path=@%-.*s", addrlen-3, addrbuf.sau.sun_path+1);
 		}
 		break;
 	case AF_INET:
@@ -446,9 +634,30 @@ int addrlen;
 #ifdef HAVE_INET_NTOP
 	case AF_INET6:
 		inet_ntop(AF_INET6, &addrbuf.sa6.sin6_addr, string_addr, sizeof(string_addr));
-		tprintf("sin6_port=htons(%u), inet_pton(AF_INET6, \"%s\", &sin6_addr), sin6_flowinfo=htonl(%u)}",
-			ntohs(addrbuf.sa6.sin6_port), string_addr, ntohl(addrbuf.sa6.sin6_flowinfo));
-		break;	
+		tprintf("sin6_port=htons(%u), inet_pton(AF_INET6, \"%s\", &sin6_addr), sin6_flowinfo=%u",
+				ntohs(addrbuf.sa6.sin6_port), string_addr,
+				addrbuf.sa6.sin6_flowinfo);
+#ifdef HAVE_SIN6_SCOPE_ID
+		{
+#if defined(HAVE_IF_INDEXTONAME) && defined(IN6_IS_ADDR_LINKLOCAL) && defined(IN6_IS_ADDR_MC_LINKLOCAL)
+		    int numericscope = 0;
+		    if (IN6_IS_ADDR_LINKLOCAL (&addrbuf.sa6.sin6_addr)
+			    || IN6_IS_ADDR_MC_LINKLOCAL (&addrbuf.sa6.sin6_addr)) {
+			char scopebuf[IFNAMSIZ + 1];
+			
+			if (if_indextoname (addrbuf.sa6.sin6_scope_id, scopebuf) == NULL)
+			    numericscope++;
+			else
+			    tprintf(", sin6_scope_id=if_nametoindex(\"%s\")", scopebuf);
+		    } else
+			numericscope++;
+		    
+		    if (numericscope)
+#endif
+			tprintf(", sin6_scope_id=%u", addrbuf.sa6.sin6_scope_id);
+		}
+#endif
+		    break;
 #endif
 #if defined(AF_IPX) && defined(linux)
 	case AF_IPX:
@@ -598,6 +807,33 @@ struct tcb *tcp;
 	return 0;
 }
 
+int
+sys_so_socket(tcp)
+struct tcb *tcp;
+{
+	if (entering(tcp)) {
+		/* not sure really what these args are... but this
+		 * is how truss prints it
+		 */
+		tprintf("%ld, %ld, %ld, ",
+		  tcp->u_arg[0], tcp->u_arg[1], tcp->u_arg[2]);
+		printpath(tcp, tcp->u_arg[3]);
+		tprintf(", %ld", tcp->u_arg[4]);
+	}
+	return 0;
+}
+
+int
+sys_so_socketpair(tcp)
+struct tcb *tcp;
+{
+	if (entering(tcp)) {
+	  	/* not sure what this arg is */
+		tprintf("0x%lx", tcp->u_arg[0]);
+	}
+	return 0;
+}
+
 int
 sys_bind(tcp)
 struct tcb *tcp;
@@ -842,7 +1078,7 @@ struct tcb *tcp;
 		else
 			tprintf("[%u, %u]", fds[0], fds[1]);
 	}
-#elif defined(SPARC) || defined(SVR4)
+#elif defined(SPARC) || defined(SVR4) || defined(FREEBSD)
 	if (exiting(tcp))
 		tprintf("[%lu, %lu]", tcp->u_rval, getrval2(tcp));
 #endif
@@ -889,12 +1125,9 @@ struct tcb *tcp;
 		else
 			tprintf(", [%u, %u]", fds[0], fds[1]);
 #endif /* LINUX */
-#ifdef SUNOS4
+#if defined(SUNOS4) || defined(SVR4) || defined(FREEBSD)
 		tprintf(", [%lu, %lu]", tcp->u_rval, getrval2(tcp));
-#endif /* SUNOS4 */
-#ifdef SVR4
-		tprintf(", [%lu, %lu]", tcp->u_rval, getrval2(tcp));
-#endif /* SVR4 */
+#endif /* SUNOS4 || SVR4 || FREEBSD */
 	}
 	return 0;
 }
@@ -925,6 +1158,13 @@ struct tcb *tcp;
 			tprintf(", ");
 			break;
 #endif
+#ifdef SOL_PACKET
+		case SOL_PACKET:
+			tprintf("SOL_PACKET, ");
+			printxval(sockpacketoptions, tcp->u_arg[2], "PACKET_???");
+			tprintf(", ");
+			break;
+#endif
 #ifdef SOL_TCP
 		case SOL_TCP:
 			tprintf("SOL_TCP, ");
@@ -937,7 +1177,7 @@ struct tcb *tcp;
 		 * etc. still need work */
 		default: 
 			/* XXX - should know socket family here */
-			printxval(protocols, tcp->u_arg[1], "IPPROTO_???");
+			printxval(socketlayers, tcp->u_arg[1], "SOL_???");
 			tprintf(", %lu, ", tcp->u_arg[2]);
 			break;
 		}
@@ -954,6 +1194,33 @@ struct tcb *tcp;
 	return 0;
 }
 
+#if defined(ICMP_FILTER)
+static void printicmpfilter(tcp, addr)
+struct tcb *tcp;
+long addr;
+{
+	struct icmp_filter	filter;
+
+	if (!addr) {
+		tprintf("NULL");
+		return;
+	}
+	if (syserror(tcp) || !verbose(tcp)) {
+		tprintf("%#lx", addr);
+		return;
+	}
+	if (umove(tcp, addr, &filter) < 0) {
+		tprintf("{...}");
+		return;
+	}
+
+	tprintf("~(");
+	if (printflags(icmpfilterflags, ~filter.data) == 0)
+		tprintf("0");
+	tprintf(")");
+}
+#endif /* ICMP_FILTER */
+
 int
 sys_setsockopt(tcp)
 struct tcb *tcp;
@@ -965,12 +1232,16 @@ struct tcb *tcp;
 			tprintf("SOL_SOCKET, ");
 			printxval(sockoptions, tcp->u_arg[2], "SO_???");
 			tprintf(", ");
+			printnum(tcp, tcp->u_arg[3], "%ld");
+			tprintf(", %lu", tcp->u_arg[4]);
 			break;
 #ifdef SOL_IP
 		case SOL_IP:
 			tprintf("SOL_IP, ");
 			printxval(sockipoptions, tcp->u_arg[2], "IP_???");
 			tprintf(", ");
+			printnum(tcp, tcp->u_arg[3], "%ld");
+			tprintf(", %lu", tcp->u_arg[4]);
 			break;
 #endif
 #ifdef SOL_IPX
@@ -978,6 +1249,18 @@ struct tcb *tcp;
 			tprintf("SOL_IPX, ");
 			printxval(sockipxoptions, tcp->u_arg[2], "IPX_???");
 			tprintf(", ");
+			printnum(tcp, tcp->u_arg[3], "%ld");
+			tprintf(", %lu", tcp->u_arg[4]);
+			break;
+#endif
+#ifdef SOL_PACKET
+		case SOL_PACKET:
+			tprintf("SOL_PACKET, ");
+			printxval(sockpacketoptions, tcp->u_arg[2], "PACKET_???");
+			tprintf(", ");
+			/* TODO: decode packate_mreq for PACKET_*_MEMBERSHIP */
+			printnum(tcp, tcp->u_arg[3], "%ld");
+			tprintf(", %lu", tcp->u_arg[4]);
 			break;
 #endif
 #ifdef SOL_TCP
@@ -985,6 +1268,26 @@ struct tcb *tcp;
 			tprintf("SOL_TCP, ");
 			printxval(socktcpoptions, tcp->u_arg[2], "TCP_???");
 			tprintf(", ");
+			printnum(tcp, tcp->u_arg[3], "%ld");
+			tprintf(", %lu", tcp->u_arg[4]);
+			break;
+#endif
+#ifdef SOL_RAW
+		case SOL_RAW:
+			tprintf("SOL_RAW, ");
+			printxval(sockrawoptions, tcp->u_arg[2], "RAW_???");
+			tprintf(", ");
+			switch (tcp->u_arg[2]) {
+#if defined(ICMP_FILTER)
+				case ICMP_FILTER:
+					printicmpfilter(tcp, tcp->u_arg[3]);
+					break;
+#endif
+				default:
+					printnum(tcp, tcp->u_arg[3], "%ld");
+					break;
+			}
+			tprintf(", %lu", tcp->u_arg[4]);
 			break;
 #endif
 
@@ -992,12 +1295,166 @@ struct tcb *tcp;
 		 * etc. still need work  */
 		default:
 			/* XXX - should know socket family here */
-			printxval(protocols, tcp->u_arg[1], "IPPROTO_???");
-			tprintf("%lu, ", tcp->u_arg[2]);
+			printxval(socketlayers, tcp->u_arg[1], "IPPROTO_???");
+			tprintf(", %lu, ", tcp->u_arg[2]);
+			printnum(tcp, tcp->u_arg[3], "%ld");
+			tprintf(", %lu", tcp->u_arg[4]);
 			break;
 		}
-		printnum(tcp, tcp->u_arg[3], "%ld");
-		tprintf(", %lu", tcp->u_arg[4]);
 	}
 	return 0;
 }
+
+#if UNIXWARE >= 7
+
+static struct xlat sock_version[] = {
+	{ __NETLIB_UW211_SVR4,	"UW211_SVR4" },
+	{ __NETLIB_UW211_XPG4,	"UW211_XPG4" },
+	{ __NETLIB_GEMINI_SVR4,	"GEMINI_SVR4" },
+	{ __NETLIB_GEMINI_XPG4,	"GEMINI_XPG4" },
+	{ __NETLIB_FP1_SVR4,	"FP1_SVR4" },
+	{ __NETLIB_FP1_XPG4,	"FP1_XPG4" },
+	{ 0,            NULL            },
+};
+
+
+int
+netlib_call(tcp, func)
+struct tcb *tcp;
+int (*func) ();
+{
+	if (entering(tcp)) {
+		int i;
+		printxval (sock_version, tcp->u_arg[0], "__NETLIB_???");
+		tprintf(", ");
+		--tcp->u_nargs;
+		for (i = 0; i < tcp->u_nargs; i++)
+			tcp->u_arg[i] = tcp->u_arg[i + 1];
+		return func (tcp);
+		
+	}
+
+	return func (tcp);
+}
+
+int
+sys_xsocket(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_socket);
+}
+
+int
+sys_xsocketpair(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_socketpair);
+}
+
+int
+sys_xbind(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_bind);
+}
+
+int
+sys_xconnect(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_connect);
+}
+
+int
+sys_xlisten(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_listen);
+}
+
+int
+sys_xaccept(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_accept);
+}
+
+int
+sys_xsendmsg(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_sendmsg);
+}
+
+int
+sys_xrecvmsg(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_recvmsg);
+}
+
+int
+sys_xgetsockaddr(tcp)
+struct tcb *tcp;
+{
+	if (entering(tcp)) {
+		printxval (sock_version, tcp->u_arg[0], "__NETLIB_???");
+		tprintf(", ");
+		if (tcp->u_arg[1] == 0) {
+			tprintf ("LOCALNAME, ");
+		}
+		else if (tcp->u_arg[1] == 1) {
+			tprintf ("REMOTENAME, ");
+		}
+		else {
+			tprintf ("%ld, ", tcp->u_arg [1]);
+		}
+		tprintf ("%ld, ", tcp->u_arg [2]);
+	} 
+	else {
+		if (tcp->u_arg[3] == 0 || syserror(tcp)) {
+			tprintf("%#lx", tcp->u_arg[3]);
+		} else {
+			printsock(tcp, tcp->u_arg[3], tcp->u_arg[4]);
+		}
+		tprintf(", ");
+		printnum(tcp, tcp->u_arg[4], "%lu");
+	}
+
+	return 0;
+
+}
+
+#if 0
+
+int
+sys_xsetsockaddr(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_setsockaddr);
+}
+
+#endif
+
+int
+sys_xgetsockopt(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_getsockopt);
+}
+
+int
+sys_xsetsockopt(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_setsockopt);
+}
+
+int
+sys_xshutdown(tcp)
+struct tcb *tcp;
+{
+	return netlib_call (tcp, sys_shutdown);
+}
+
+#endif
