@@ -31,7 +31,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: process.c,v 1.16 1999/12/24 07:22:25 drepper Exp $
+ *	$Id: process.c,v 1.28 2001/03/28 20:29:17 wichert Exp $
  */
 
 #include "defs.h"
@@ -49,10 +49,23 @@
 #include <machine/reg.h>
 #endif /* SUNOS4 */
 
-#if HAVE_LINUX_PTRACE_H
-#undef PTRACE_SYSCALL
-#include <linux/ptrace.h>
-#endif 
+#ifdef FREEBSD
+#include <sys/ptrace.h>
+#endif
+
+#if HAVE_ASM_REG_H
+#ifdef SPARC
+#  define fpq kernel_fpq
+#  define fq kernel_fq
+#  define fpu kernel_fpu
+#endif
+#include <asm/reg.h>
+#ifdef SPARC
+#  undef fpq
+#  undef fq
+#  undef fpu 
+#endif
+#endif /* HAVE_ASM_REG_H */
 
 #ifdef HAVE_SYS_REG_H
 # include <sys/reg.h>
@@ -62,14 +75,23 @@
 #ifndef PTRACE_POKEUSR
 # define PTRACE_POKEUSR PTRACE_POKEUSER
 #endif
+#elif defined(HAVE_LINUX_PTRACE_H)
+#undef PTRACE_SYSCALL
+#include <linux/ptrace.h>
 #endif
 
+
 #ifdef LINUX
 #include <asm/posix_types.h>
 #undef GETGROUPS_T
 #define GETGROUPS_T __kernel_gid_t
 #endif /* LINUX */
 
+#if defined(LINUX) && defined(IA64)
+# include <asm/ptrace_offsets.h>
+# include <asm/rse.h>
+#endif
+
 #ifdef HAVE_PRCTL
 #include <sys/prctl.h>
 #endif
@@ -140,10 +162,45 @@ static struct xlat prctl_options[] = {
 #endif
 #ifdef PR_GET_PDEATHSIG
 	{ PR_GET_PDEATHSIG,	"PR_GET_PDEATHSIG"	},
+#endif
+#ifdef PR_GET_UNALIGN
+	{ PR_GET_UNALIGN,	"PR_GET_UNALIGN"	},
+#endif
+#ifdef PR_SET_UNALIGN
+	{ PR_SET_UNALIGN,	"PR_SET_UNALIGN"	},
+#endif
+#ifdef PR_GET_KEEPCAPS
+	{ PR_GET_KEEPCAPS,	"PR_GET_KEEP_CAPS"	},
+#endif
+#ifdef PR_SET_KEEPCAPS
+	{ PR_SET_KEEPCAPS,	"PR_SET_KEEP_CAPS"	},
 #endif
 	{ 0,			NULL			},
 };
 
+
+const char *
+unalignctl_string (unsigned int ctl)
+{
+	static char buf[16];
+
+	switch (ctl) {
+#ifdef PR_UNALIGN_NOPRINT
+	      case PR_UNALIGN_NOPRINT:
+		return "NOPRINT";
+#endif
+#ifdef PR_UNALIGN_SIGBUS
+	      case PR_UNALIGN_SIGBUS:
+		return "SIGBUS";
+#endif
+	      default:
+		break;
+	}
+	sprintf(buf, "%x", ctl);
+	return buf;
+}
+
+
 int
 sys_prctl(tcp)
 struct tcb *tcp;
@@ -160,6 +217,16 @@ struct tcb *tcp;
 #ifdef PR_SET_DEATHSIG
 		case PR_GET_PDEATHSIG:
 			break;
+#endif
+#ifdef PR_SET_UNALIGN
+		case PR_SET_UNALIGN:
+			tprintf(", %s", unalignctl_string(tcp->u_arg[1]));
+			break;
+#endif
+#ifdef PR_GET_UNALIGN
+		case PR_GET_UNALIGN:
+			tprintf(", %#lx", tcp->u_arg[1]);
+			break;
 #endif
 		default:
 			for (i = 1; i < tcp->u_nargs; i++)
@@ -173,6 +240,20 @@ struct tcb *tcp;
 			for (i=1; i<tcp->u_nargs; i++)
 				tprintf(", %#lx", tcp->u_arg[i]);
 			break;
+#endif
+#ifdef PR_SET_UNALIGN
+		case PR_SET_UNALIGN:
+			break;
+#endif
+#ifdef PR_GET_UNALIGN
+		case PR_GET_UNALIGN:
+		{
+			int ctl;
+
+			umove(tcp, tcp->u_arg[1], &ctl);
+			tcp->auxstr = unalignctl_string(ctl);
+			return RVAL_STR;
+		}
 #endif
 		default:
 			break;
@@ -270,7 +351,7 @@ struct tcb *tcp;
 	return 0;
 }
 
-#ifdef SVR4
+#ifdef USE_PROCFS
 
 int
 sys_fork(tcp)
@@ -309,12 +390,13 @@ struct tcb *tcp;
 			fprintf(stderr, "sys_fork: tcb table full\n");
 			return 0;
 		}
-		proc_open(tcpchild, 1);
+		if (proc_open(tcpchild, 2) < 0)
+		  	droptcb(tcpchild);
 	}
 	return 0;
 }
 
-#else /* !SVR4 */
+#else /* !USE_PROCFS */
 
 #ifdef LINUX
 
@@ -343,6 +425,18 @@ static struct xlat clone_flags[] = {
     { 0,		NULL		},
 };
 
+int
+sys_clone(tcp)
+struct tcb *tcp;
+{
+	if (exiting(tcp)) {
+		tprintf("child_stack=%#lx, flags=", tcp->u_arg[1]);
+		if (printflags(clone_flags, tcp->u_arg[0]) == 0)
+			tprintf("0");
+	}
+	return 0;
+}
+
 #endif
 
 int
@@ -354,6 +448,107 @@ struct tcb *tcp;
 	return 0;
 }
 
+int
+change_syscall(tcp, new)
+struct tcb *tcp;
+int new;
+{
+#if defined(LINUX)
+#if defined(I386)
+	/* Attempt to make vfork into fork, which we can follow. */
+	if (ptrace(PTRACE_POKEUSER, tcp->pid, (char*)(ORIG_EAX * 4), new) < 0) 
+		return -1;
+	return 0;
+#elif defined(POWERPC)
+	if (ptrace(PTRACE_POKEUSER, tcp->pid, (CHAR*)(4*PT_R0), new) < 0)
+		return -1;
+#elif defined(S390)
+	long	pc;
+	if (upeek(tcp->pid, PT_PSWADDR,&pc)<0)
+	    	return -1;
+	if (ptrace(PTRACE_POKETEXT, tcp->pid, (char*)(pc-4), new)<0)
+	    	return -1;
+	return 0;
+#elif defined(M68K)
+	if (ptrace(PTRACE_POKEUSER, tcp->pid, (char*)(4*PT_ORIG_D0), new)<0)
+	    	return -1;
+	return 0;
+#elif defined(SPARC)
+	struct regs regs;
+	if (ptrace(PTRACE_GETREGS, tcp->pid, (char*)&regs, 0)<0)
+		return -1;
+	regs.r_g1=new;
+	if (ptrace(PTRACE_SETREGS, tcp->pid, (char*)&regs, 0)<0)
+	    	return -1;
+	return 0;
+#elif defined(MIPS)
+	if (ptrace(PTRACE_POKEUSER, tcp->pid, (char*)(REG_V0), new)<0)
+	    	return -1;
+	return 0;
+#elif defined(ALPHA)
+	if (ptrace(PTRACE_POKEUSER, tcp->pid, (char*)(REG_A3), new)<0)
+	    	return -1;
+	return 0;
+#elif defined(HPPA)
+	if (ptrace(PTRACE_POKEUSER, tcp->pid, (char*)(PT_GR20), new)<0)
+	    	return -1;
+	return 0;
+#else
+#warning Do not know how to handle change_syscall for this architecture
+#endif /* architecture */
+#endif /* LINUX */
+	return -1;
+}
+
+int
+setarg(tcp, argnum)
+	struct tcb *tcp;
+	int argnum;
+{
+#if defined (IA64)
+	{
+		unsigned long *bsp, *ap;
+
+		if (upeek(tcp->pid, PT_AR_BSP, (long *) &bsp) , 0)
+			return -1;
+
+		ap = ia64_rse_skip_regs(bsp, argnum);
+		errno = 0;
+		ptrace(PTRACE_POKEDATA, tcp->pid, ap, tcp->u_arg[argnum]);
+		if (errno)
+			return -1;
+
+	}
+#elif defined(I386)
+	{
+		ptrace(PTRACE_POKEUSER, tcp->pid, (char*)(4*argnum), tcp->u_arg[argnum]);
+		if (errno)
+			return -1;
+	}
+#elif defined(MIPS)
+	{
+		errno = 0;
+		if (argnum < 4)
+			ptrace(PTRACE_POKEUSER, tcp->pid,
+			       (char*)(REG_A0 + argnum), tcp->u_arg[argnum]);
+		else {
+			unsigned long *sp;
+
+			if (upeek(tcp->pid, REG_SP, (long *) &sp) , 0)
+				return -1;
+
+			ptrace(PTRACE_POKEDATA, tcp->pid,
+			       (char*)(sp + argnum - 4), tcp->u_arg[argnum]);
+		}
+		if (errno)
+			return -1;
+	}
+#else
+# warning Sorry, setargs not implemented for this architecture.
+#endif
+	return 0;
+}
+
 #ifdef SYS_clone
 int
 internal_clone(tcp)
@@ -361,18 +556,8 @@ struct tcb *tcp;
 {
 	struct tcb *tcpchild;
 	int pid;
-	int dont_follow = 0;
 	if (entering(tcp)) {
-		tprintf("fn=%#lx, child_stack=%#lx, flags=",
-				tcp->u_arg[0], tcp->u_arg[1]);
-		if (printflags(clone_flags, tcp->u_arg[2]) == 0)
-			tprintf("0");
-		tprintf(", args=%#lx", tcp->u_arg[3]);
-
-		/* For now we don't follow clone yet.. we're just preparing the code */
-		dont_follow = 1;
-
-		if (!followfork || dont_follow)
+		if (!followfork)
 			return 0;
 		if (nprocs == MAX_PROCS) {
 			tcp->flags &= ~TCB_FOLLOWFORK;
@@ -381,19 +566,18 @@ struct tcb *tcp;
 		}
 		tcp->flags |= TCB_FOLLOWFORK;
 
-		/* XXX 
-		 * We will take the simple approach and add CLONE_PTRACE to the clone
-		 * options. This only works on Linux 2.2.x and later. This means that
-		 * we break all programs using clone on older kernels..
-		 * We should try to fallback to the bpt-trick if this fails, but right
-		 * now we don't.
-		 */
 
-		/* TODO: actually change the flags */
+		if (setbpt(tcp) < 0)
+			return 0;
 	} else {
+		int bpt = tcp->flags & TCB_BPTSET;
+
 		if (!(tcp->flags & TCB_FOLLOWFORK))
 			return 0;
 
+		if (bpt)
+			clearbpt(tcp);
+
 		if (syserror(tcp))
 			return 0;
 
@@ -404,10 +588,21 @@ struct tcb *tcp;
 			return 0;
 		}
 
-		/* For fork we need to re-attach, but thanks to CLONE_PTRACE we're
-		 * already attached.
-		 */
+		/* Attach to the new child */
+		if (ptrace(PTRACE_ATTACH, pid, (char *) 1, 0) < 0) {
+			perror("PTRACE_ATTACH");
+			fprintf(stderr, "Too late?\n");
+			droptcb(tcpchild);
+			return 0;
+		}
+
 		tcpchild->flags |= TCB_ATTACHED;
+		if (bpt) {
+			tcpchild->flags |= TCB_BPTSET;
+			tcpchild->baddr = tcp->baddr;
+			memcpy(tcpchild->inst, tcp->inst,
+				sizeof tcpchild->inst);
+		}
 		newoutf(tcpchild);
 		tcp->nchildren++;
 		if (!qflag)
@@ -427,30 +622,10 @@ struct tcb *tcp;
 
 #ifdef SYS_vfork
 	if (tcp->scno == SYS_vfork) {
-#if defined(I386) && defined(LINUX)
 		/* Attempt to make vfork into fork, which we can follow. */
 		if (!followvfork || 
-		    ptrace(PTRACE_POKEUSR, tcp->pid, 
-			   (void *)(ORIG_EAX * 4), SYS_fork) < 0) 
+		    change_syscall(tcp, SYS_fork) < 0)
 			dont_follow = 1;
-		
-#else
-		dont_follow = 1;
-#endif
-	}
-#endif
-#ifdef SYS_clone
-	/* clone can do many things, not all of which we know how to handle.
-	   Don't do much for now. */
-	if (tcp->scno == SYS_clone) {
-	    	if (entering(tcp)) {
-			tprintf("fn=%#lx, child_stack=%#lx, flags=",
-					tcp->u_arg[0], tcp->u_arg[1]);
-			if (printflags(clone_flags, tcp->u_arg[2]) == 0)
-			    tprintf("0");
-			tprintf(", args=%#lx", tcp->u_arg[3]);
-		}
-		dont_follow = 1;
 	}
 #endif
 	if (entering(tcp)) {
@@ -483,6 +658,24 @@ struct tcb *tcp;
 			return 0;
 		}
 #ifdef LINUX
+#ifdef HPPA
+		/* The child must have run before it can be attached. */
+		/* This must be a bug in the parisc kernel, but I havn't
+		 * identified it yet.  Seems to be an issue associated
+		 * with attaching to a process (which sends it a signal)
+		 * before that process has ever been scheduled.  When
+		 * debugging, I started seeing crashes in
+		 * arch/parisc/kernel/signal.c:do_signal(), apparently
+		 * caused by r8 getting corrupt over the dequeue_signal()
+		 * call.  Didn't make much sense though...
+		 */
+		{
+			struct timeval tv;
+			tv.tv_sec = 0;
+			tv.tv_usec = 10000;
+			select(0, NULL, NULL, NULL, &tv);
+		}
+#endif
 		if (ptrace(PTRACE_ATTACH, pid, (char *) 1, 0) < 0) {
 			perror("PTRACE_ATTACH");
 			fprintf(stderr, "Too late?\n");
@@ -538,9 +731,9 @@ struct tcb *tcp;
 	return 0;
 }
 
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 
-#if defined(SUNOS4) || defined(LINUX)
+#if defined(SUNOS4) || defined(LINUX) || defined(FREEBSD)
 
 int
 sys_vfork(tcp)
@@ -551,7 +744,7 @@ struct tcb *tcp;
 	return 0;
 }
 
-#endif /* SUNOS4 || LINUX */
+#endif /* SUNOS4 || LINUX || FREEBSD */
 
 #ifndef LINUX
 
@@ -697,7 +890,7 @@ struct tcb *tcp;
 	return 0;
 }
 
-#ifdef LINUX
+#if defined(LINUX) || defined(FREEBSD)
 int
 sys_setresuid(tcp)
      struct tcb *tcp;
@@ -723,7 +916,7 @@ sys_setresgid(tcp)
 	return 0;
 }
 
-#endif /* LINUX */
+#endif /* LINUX || FREEBSD */
 
 int
 sys_setgroups(tcp)
@@ -985,7 +1178,7 @@ struct tcb *tcp;
 		}
 	}
 #ifdef LINUX
-#if defined(ALPHA) || defined(SPARC) || defined(POWERPC)
+#if defined(ALPHA) || defined(SPARC) || defined(POWERPC) || defined(HPPA)
 	tcp->flags |= TCB_WAITEXECVE;
 #endif /* ALPHA || SPARC || POWERPC */
 #endif /* LINUX */
@@ -1145,6 +1338,25 @@ struct tcb *tcp;
 
 #endif /* SVR4 */
 
+#ifdef FREEBSD
+int
+sys_wait(tcp)
+struct tcb *tcp;
+{
+	int status;
+	
+	if (exiting(tcp)) {
+		if (!syserror(tcp)) {
+			if (umove(tcp, tcp->u_arg[0], &status) < 0)
+				tprintf("%#lx", tcp->u_arg[0]);
+			else
+				printstatus(status);
+		}
+	}
+	return 0;
+}
+#endif
+
 int
 sys_waitpid(tcp)
 struct tcb *tcp;
@@ -1473,6 +1685,7 @@ struct tcb *tcp;
 #ifndef SVR4
 
 static struct xlat ptrace_cmds[] = {
+#ifndef FREEBSD	
 	{ PTRACE_TRACEME,	"PTRACE_TRACEME"	},
 	{ PTRACE_PEEKTEXT,	"PTRACE_PEEKTEXT",	},
 	{ PTRACE_PEEKDATA,	"PTRACE_PEEKDATA",	},
@@ -1518,9 +1731,29 @@ static struct xlat ptrace_cmds[] = {
 #endif /* !I386 */
 	{ PTRACE_GETUCODE,	"PTRACE_GETUCODE"	},
 #endif /* SUNOS4 */
+#else /* FREEBSD */
+	{ PT_TRACE_ME,		"PT_TRACE_ME"		},
+	{ PT_READ_I,		"PT_READ_I"		},
+	{ PT_READ_D,		"PT_READ_D"		},
+	{ PT_WRITE_I,		"PT_WRITE_I"		},
+	{ PT_WRITE_D,		"PT_WRITE_D"		},
+	{ PT_READ_U,		"PT_WRITE_U"		},
+	{ PT_CONTINUE,		"PT_CONTINUE"		},
+	{ PT_KILL,		"PT_KILL"		},
+	{ PT_STEP,		"PT_STEP"		},
+	{ PT_ATTACH,		"PT_ATTACH"		},
+	{ PT_DETACH,		"PT_DETACH"		},
+	{ PT_GETREGS,		"PT_GETREGS"		},
+	{ PT_SETREGS,		"PT_SETREGS"		},
+	{ PT_GETFPREGS,		"PT_GETFPREGS"		},
+	{ PT_SETFPREGS,		"PT_SETFPREGS"		},
+	{ PT_GETDBREGS,		"PT_GETDBREGS"		},
+	{ PT_SETDBREGS,		"PT_SETDBREGS"		},
+#endif /* FREEBSD */
 	{ 0,			NULL			},
 };
 
+#ifndef FREEBSD
 #ifndef SUNOS4_KERNEL_ARCH_KLUDGE
 static
 #endif /* !SUNOS4_KERNEL_ARCH_KLUDGE */
@@ -1601,6 +1834,8 @@ struct xlat struct_user_offsets[] = {
 #endif
 #if defined(SPARC)
 	/* XXX No support for these offsets yet. */
+#elif defined(HPPA)
+	/* XXX No support for these offsets yet. */
 #elif defined(POWERPC)
 	{ 4*PT_R0,		"4*PT_R0"				},
 	{ 4*PT_R1,		"4*PT_R1"				},
@@ -1710,6 +1945,76 @@ struct xlat struct_user_offsets[] = {
 	{ 63,			"fp31"					},
 	{ 64,			"pc"					},
 #else /* !ALPHA */
+#ifdef IA64
+	{ PT_F32, "f32" }, { PT_F33, "f33" }, { PT_F34, "f34" },
+	{ PT_F35, "f35" }, { PT_F36, "f36" }, { PT_F37, "f37" },
+	{ PT_F38, "f38" }, { PT_F39, "f39" }, { PT_F40, "f40" },
+	{ PT_F41, "f41" }, { PT_F42, "f42" }, { PT_F43, "f43" },
+	{ PT_F44, "f44" }, { PT_F45, "f45" }, { PT_F46, "f46" },
+	{ PT_F47, "f47" }, { PT_F48, "f48" }, { PT_F49, "f49" },
+	{ PT_F50, "f50" }, { PT_F51, "f51" }, { PT_F52, "f52" },
+	{ PT_F53, "f53" }, { PT_F54, "f54" }, { PT_F55, "f55" },
+	{ PT_F56, "f56" }, { PT_F57, "f57" }, { PT_F58, "f58" },
+	{ PT_F59, "f59" }, { PT_F60, "f60" }, { PT_F61, "f61" },
+	{ PT_F62, "f62" }, { PT_F63, "f63" }, { PT_F64, "f64" },
+	{ PT_F65, "f65" }, { PT_F66, "f66" }, { PT_F67, "f67" },
+	{ PT_F68, "f68" }, { PT_F69, "f69" }, { PT_F70, "f70" },
+	{ PT_F71, "f71" }, { PT_F72, "f72" }, { PT_F73, "f73" },
+	{ PT_F74, "f74" }, { PT_F75, "f75" }, { PT_F76, "f76" },
+	{ PT_F77, "f77" }, { PT_F78, "f78" }, { PT_F79, "f79" },
+	{ PT_F80, "f80" }, { PT_F81, "f81" }, { PT_F82, "f82" },
+	{ PT_F83, "f83" }, { PT_F84, "f84" }, { PT_F85, "f85" },
+	{ PT_F86, "f86" }, { PT_F87, "f87" }, { PT_F88, "f88" },
+	{ PT_F89, "f89" }, { PT_F90, "f90" }, { PT_F91, "f91" },
+	{ PT_F92, "f92" }, { PT_F93, "f93" }, { PT_F94, "f94" },
+	{ PT_F95, "f95" }, { PT_F96, "f96" }, { PT_F97, "f97" },
+	{ PT_F98, "f98" }, { PT_F99, "f99" }, { PT_F100, "f100" },
+	{ PT_F101, "f101" }, { PT_F102, "f102" }, { PT_F103, "f103" },
+	{ PT_F104, "f104" }, { PT_F105, "f105" }, { PT_F106, "f106" },
+	{ PT_F107, "f107" }, { PT_F108, "f108" }, { PT_F109, "f109" },
+	{ PT_F110, "f110" }, { PT_F111, "f111" }, { PT_F112, "f112" },
+	{ PT_F113, "f113" }, { PT_F114, "f114" }, { PT_F115, "f115" },
+	{ PT_F116, "f116" }, { PT_F117, "f117" }, { PT_F118, "f118" },
+	{ PT_F119, "f119" }, { PT_F120, "f120" }, { PT_F121, "f121" },
+	{ PT_F122, "f122" }, { PT_F123, "f123" }, { PT_F124, "f124" },
+	{ PT_F125, "f125" }, { PT_F126, "f126" }, { PT_F127, "f127" },
+	/* switch stack: */
+	{ PT_F2, "f2" }, { PT_F3, "f3" }, { PT_F4, "f4" },
+	{ PT_F5, "f5" }, { PT_F10, "f10" }, { PT_F11, "f11" },
+	{ PT_F12, "f12" }, { PT_F13, "f13" }, { PT_F14, "f14" },
+	{ PT_F15, "f15" }, { PT_F16, "f16" }, { PT_F17, "f17" },
+	{ PT_F18, "f18" }, { PT_F19, "f19" }, { PT_F20, "f20" },
+	{ PT_F21, "f21" }, { PT_F22, "f22" }, { PT_F23, "f23" },
+	{ PT_F24, "f24" }, { PT_F25, "f25" }, { PT_F26, "f26" },
+	{ PT_F27, "f27" }, { PT_F28, "f28" }, { PT_F29, "f29" },
+	{ PT_F30, "f30" }, { PT_F31, "f31" }, { PT_R4, "r4" },
+	{ PT_R5, "r5" }, { PT_R6, "r6" }, { PT_R7, "r7" },
+	{ PT_K_B0, "kb0" },
+	{ PT_B1, "b1" }, { PT_B2, "b2" }, { PT_B3, "b3" },
+	{ PT_B4, "b4" }, { PT_B5, "b5" },
+	{ PT_K_AR_PFS, "kar.pfs" },
+	{ PT_AR_LC, "ar.lc" }, { PT_K_AR_UNAT, "kar.unat" },
+	{ PT_K_AR_RNAT, "kar.rnat" }, { PT_K_AR_BSPSTORE, "kar.bspstore" },
+	{ PT_K_PR, "k.pr" },
+	/* pt_regs */
+	{ PT_CR_IPSR, "cr.ipsr" }, { PT_CR_IIP, "cr.iip" },
+	{ PT_CR_IFS, "cr.ifs" }, { PT_AR_UNAT, "ar.unat" },
+	{ PT_AR_PFS, "ar.pfs" }, { PT_AR_RSC, "ar.rsc" },
+	{ PT_AR_RNAT, "ar.rnat" }, { PT_AR_BSPSTORE, "ar.bspstore" },
+	{ PT_PR, "pr" }, { PT_B6, "b6" }, { PT_AR_BSP, "ar.bsp" },
+	{ PT_R1, "r1" }, { PT_R2, "r2" }, { PT_R3, "r3" },
+	{ PT_R12, "r12" }, { PT_R13, "r13" }, { PT_R14, "r14" },
+	{ PT_R15, "r15" }, { PT_R8, "r8" }, { PT_R9, "r9" },
+	{ PT_R10, "r10" }, { PT_R11, "r11" }, { PT_R16, "r16" },
+	{ PT_R17, "r17" }, { PT_R18, "r18" }, { PT_R19, "r19" },
+	{ PT_R20, "r20" }, { PT_R21, "r21" }, { PT_R22, "r22" },
+	{ PT_R23, "r23" }, { PT_R24, "r24" }, { PT_R25, "r25" },
+	{ PT_R26, "r26" }, { PT_R27, "r27" }, { PT_R28, "r28" },
+	{ PT_R29, "r29" }, { PT_R30, "r30" }, { PT_R31, "r31" },
+	{ PT_AR_CCV, "ar.ccv" }, { PT_AR_FPSR, "ar.fpsr" },
+	{ PT_B0, "b0" }, { PT_B7, "b7" }, { PT_F6, "f6" },
+	{ PT_F7, "f7" }, { PT_F8, "f8" }, { PT_F9, "f9" },
+#else /* !IA64 */
 #ifdef I386
 	{ 4*EBX,		"4*EBX"					},
 	{ 4*ECX,		"4*ECX"					},
@@ -1751,10 +2056,7 @@ struct xlat struct_user_offsets[] = {
 	{ 4*PT_PC,		"4*PT_PC"				},
 #endif /* M68K */
 #endif /* !I386 */
-#ifdef S390
-	{ uoff(u_fpvalid),	"offsetof(struct user, u_fpvalid)"	},
-#endif
-#ifndef MIPS
+#if !defined(S390) && !defined(MIPS)
 	{ uoff(u_fpvalid),	"offsetof(struct user, u_fpvalid)"	},
 #endif
 #ifdef I386
@@ -1770,7 +2072,7 @@ struct xlat struct_user_offsets[] = {
 	{ uoff(start_code),	"offsetof(struct user, start_code)"	},
 	{ uoff(start_stack),	"offsetof(struct user, start_stack)"	},
 	{ uoff(signal),		"offsetof(struct user, signal)"		},
-#ifndef MIPS
+#if !defined(S390) && !defined(MIPS)
 	{ uoff(reserved),	"offsetof(struct user, reserved)"	},
 #endif
 	{ uoff(u_ar0),		"offsetof(struct user, u_ar0)"		},
@@ -1782,6 +2084,7 @@ struct xlat struct_user_offsets[] = {
 #ifdef I386
 	{ uoff(u_debugreg),	"offsetof(struct user, u_debugreg)"	},
 #endif /* I386 */
+#endif /* !IA64 */
 #endif /* !ALPHA */
 #endif /* !POWERPC/!SPARC */
 #endif /* LINUX */
@@ -1832,9 +2135,12 @@ struct xlat struct_user_offsets[] = {
 	{ uoff(u_exdata.ux_shell[0]),"offsetof(struct user, u_exdata.ux_shell[0])"},
 	{ uoff(u_lofault),	"offsetof(struct user, u_lofault)"	},
 #endif /* SUNOS4 */
+#ifndef HPPA
 	{ sizeof(struct user),	"sizeof(struct user)"			},
+#endif
 	{ 0,			NULL					},
 };
+#endif
 
 int
 sys_ptrace(tcp)
@@ -1846,10 +2152,15 @@ struct tcb *tcp;
 
 	cmd = xlookup(ptrace_cmds, tcp->u_arg[0]);
 	if (!cmd)
+#ifndef FREEBSD
 		cmd = "PTRACE_???";
+#else
+		cmd = "PT_???";
+#endif		
 	if (entering(tcp)) {
 		tprintf("%s, %lu, ", cmd, tcp->u_arg[1]);
 		addr = tcp->u_arg[2];
+#ifndef FREEBSD
 		if (tcp->u_arg[0] == PTRACE_PEEKUSER
 			|| tcp->u_arg[0] == PTRACE_POKEUSER) {
 			for (x = struct_user_offsets; x->str; x++) {
@@ -1866,6 +2177,7 @@ struct tcb *tcp;
 				tprintf("%s, ", x->str);
 		}
 		else
+#endif
 			tprintf("%#lx, ", tcp->u_arg[2]);
 #ifdef LINUX
 		switch (tcp->u_arg[0]) {
@@ -1910,6 +2222,10 @@ struct tcb *tcp;
 		}
 	}
 #endif /* SUNOS4 */
+#ifdef FREEBSD
+		tprintf("%lu", tcp->u_arg[3]);
+	}
+#endif /* FREEBSD */
 	return 0;
 }
 
