@@ -30,7 +30,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: signal.c,v 1.18 1999/12/23 14:20:15 wichert Exp $
+ *	$Id: signal.c,v 1.28 2001/03/28 20:29:17 wichert Exp $
  */
 
 #include "defs.h"
@@ -43,11 +43,6 @@
 #include <sys/ucontext.h>
 #endif /* SVR4 */
 
-#if HAVE_LINUX_PTRACE_H
-#undef PTRACE_SYSCALL
-#include <linux/ptrace.h>
-#endif 
-
 #ifdef HAVE_SYS_REG_H
 # include <sys/reg.h>
 #ifndef PTRACE_PEEKUSR
@@ -56,19 +51,41 @@
 #ifndef PTRACE_POKEUSR
 # define PTRACE_POKEUSR PTRACE_POKEUSER
 #endif
+#elif defined(HAVE_LINUX_PTRACE_H)
+#undef PTRACE_SYSCALL
+#include <linux/ptrace.h>
 #endif
 
+
 #ifdef LINUX
 
-#ifdef HAVE_ASM_SIGCONTEXT_H
-#include <asm/sigcontext.h>
+#ifdef IA64
+# include <asm/ptrace_offsets.h>
+#endif /* !IA64 */
+
+#if HAVE_ASM_REG_H
 #ifdef SPARC
+#  define fpq kernel_fpq
+#  define fq kernel_fq
+#  define fpu kernel_fpu
+#endif
 #include <asm/reg.h>
+#ifdef SPARC
+#  undef fpq
+#  undef fq
+#  undef fpu 
+#endif
+
+#endif /* HAVE_ASM_REG_H */
+#ifdef HAVE_ASM_SIGCONTEXT_H
+#ifdef SPARC
 typedef struct {
 	struct regs		si_regs;
 	int			si_mask;
 } m_siginfo_t;
-#endif
+#else
+#include <asm/sigcontext.h>
+#endif /* SPARC */
 #else /* !HAVE_ASM_SIGCONTEXT_H */
 #ifdef I386
 struct sigcontext_struct {
@@ -143,7 +160,7 @@ int nsignals2 = sizeof signalent2 / sizeof signalent2[0];
 char **signalent;
 int nsignals;
 
-#ifdef SUNOS4
+#if defined(SUNOS4) || defined(FREEBSD)
 
 static struct xlat sigvec_flags[] = {
 	{ SV_ONSTACK,	"SV_ONSTACK"	},
@@ -153,11 +170,14 @@ static struct xlat sigvec_flags[] = {
 	{ 0,		NULL		},
 };
 
-#endif /* SUNOS4 */
+#endif /* SUNOS4 || FREEBSD */
 
 #ifdef HAVE_SIGACTION
 
 static struct xlat sigact_flags[] = {
+#ifdef SA_RESTORER
+	{ SA_RESTORER,	"SA_RESTORER"	},
+#endif
 #ifdef SA_STACK
 	{ SA_STACK,	"SA_STACK"	},
 #endif
@@ -251,7 +271,7 @@ sigset_t *s;
 static int
 copy_sigset_len(tcp, addr, s, len)
 struct tcb *tcp;
-int addr;
+long addr;
 sigset_t *s;
 int len;
 {
@@ -310,7 +330,18 @@ int rt; /* set might include realtime sigs */
 	*s++ = '[';
 	for (i = 1; i < maxsigs; i++) {
 		if (sigismember(mask, i) == 1) {
-			sprintf(s, format, signame(i) + 3); s += strlen(s);
+			/* real-time signals on solaris don't have
+			 * signalent entries
+			 */
+			if (i < nsignals) {
+				sprintf(s, format, signalent[i] + 3);
+			}
+			else {
+				char tsig[32];
+				sprintf(tsig, "%u", i);
+				sprintf(s, format, tsig);
+			}
+			s += strlen(s);
 			format = " %s";
 		}
 	}
@@ -423,7 +454,7 @@ int sig;
 #endif /* !SVR4 */
 }
 
-#if defined(SUNOS4)
+#if defined(SUNOS4) || defined(FREEBSD)
 
 int
 sys_sigvec(tcp)
@@ -528,7 +559,7 @@ struct tcb *tcp;
 	return 0;
 }
 
-#endif /* SUNOS4 */
+#endif /* SUNOS4 || FREEBSD */
 
 #ifndef SVR4
 
@@ -540,12 +571,14 @@ struct tcb *tcp;
 		sigset_t sigm;
 		long_to_sigset(tcp->u_arg[0], &sigm);
 		printsigmask(&sigm, 0);
+#ifndef USE_PROCFS
 		if ((tcp->u_arg[0] & sigmask(SIGTRAP))) {
 			/* Mark attempt to block SIGTRAP */
 			tcp->flags |= TCB_SIGTRAPPED;
 			/* Send unblockable signal */
 			kill(tcp->pid, SIGSTOP);
 		}
+#endif /* !USE_PROCFS */		
 	}
 	else if (!syserror(tcp)) {
 		sigset_t sigm;
@@ -587,8 +620,8 @@ sys_sigaction(tcp)
 struct tcb *tcp;
 {
 	long addr;
-	sigset_t sigset;
 #ifdef LINUX
+	sigset_t sigset;
 	struct old_sigaction sa;
 #else
 	struct sigaction sa;
@@ -616,21 +649,21 @@ struct tcb *tcp;
 			tprintf("{SIG_DFL}");
 			break;
 		case (long) SIG_IGN:
-#ifndef SVR4
+#ifndef USE_PROCFS
 			if (tcp->u_arg[0] == SIGTRAP) {
 				tcp->flags |= TCB_SIGTRAPPED;
 				kill(tcp->pid, SIGSTOP);
 			}
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 			tprintf("{SIG_IGN}");
 			break;
 		default:
-#ifndef SVR4
+#ifndef USE_PROCFS
 			if (tcp->u_arg[0] == SIGTRAP) {
 				tcp->flags |= TCB_SIGTRAPPED;
 				kill(tcp->pid, SIGSTOP);
 			}
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 			tprintf("{%#lx, ", (long) sa.SA_HANDLER);
 #ifndef LINUX
 			printsigmask (&sa.sa_mask, 0);
@@ -659,6 +692,7 @@ struct tcb *tcp;
 {
 	if (entering(tcp)) {
 		printsignal(tcp->u_arg[0]);
+		tprintf(", ");
 		switch (tcp->u_arg[1]) {
 		case (int) SIG_ERR:
 			tprintf("SIG_ERR");
@@ -667,23 +701,46 @@ struct tcb *tcp;
 			tprintf("SIG_DFL");
 			break;
 		case (int) SIG_IGN:
-#ifndef SVR4
+#ifndef USE_PROCFS
 			if (tcp->u_arg[0] == SIGTRAP) {
 				tcp->flags |= TCB_SIGTRAPPED;
 				kill(tcp->pid, SIGSTOP);
 			}
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 			tprintf("SIG_IGN");
 			break;
 		default:
-#ifndef SVR4
+#ifndef USE_PROCFS
 			if (tcp->u_arg[0] == SIGTRAP) {
 				tcp->flags |= TCB_SIGTRAPPED;
 				kill(tcp->pid, SIGSTOP);
 			}
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 			tprintf("%#lx", tcp->u_arg[1]);
 		}
+		return 0;
+	}
+	else {
+		switch (tcp->u_rval) {
+		    case (int) SIG_ERR:
+			tcp->auxstr = "SIG_ERR"; break;
+		    case (int) SIG_DFL:
+			tcp->auxstr = "SIG_DFL"; break;
+		    case (int) SIG_IGN:
+			tcp->auxstr = "SIG_IGN"; break;
+		    default:
+			tcp->auxstr = NULL;
+		}
+		return RVAL_HEX | RVAL_STR;
+	}
+}
+
+int
+sys_sighold(tcp)
+struct tcb *tcp;
+{
+	if (entering(tcp)) {
+		printsignal(tcp->u_arg[0]);
 	}
 	return 0;
 }
@@ -741,6 +798,33 @@ struct tcb *tcp;
 	}
 	return 0;
 #else /* !I386 */
+#ifdef IA64
+	struct sigcontext sc;
+	long sp;
+
+	if (entering(tcp)) {
+		/* offset of sigcontext in the kernel's sigframe structure: */
+#		define SIGFRAME_SC_OFFSET	0x90
+		tcp->u_arg[0] = 0;
+		if (upeek(tcp->pid, PT_R12, &sp) < 0)
+			return 0;
+		if (umove(tcp, sp + 16 + SIGFRAME_SC_OFFSET, &sc) < 0)
+			return 0;
+		tcp->u_arg[0] = 1;
+		memcpy(tcp->u_arg + 1, &sc.sc_mask, sizeof(sc.sc_mask));
+	}
+	else {
+		sigset_t sigm;
+
+		memcpy(&sigm, tcp->u_arg + 1, sizeof (sigm));
+		tcp->u_rval = tcp->u_error = 0;
+		if (tcp->u_arg[0] == 0)
+			return 0;
+		tcp->auxstr = sprintsigmask("mask now ", &sigm, 0);
+		return RVAL_NONE | RVAL_STR;
+	}
+	return 0;
+#else /* !IA64 */
 #ifdef POWERPC
        long esp;
        struct sigcontext_struct sc;
@@ -862,11 +946,17 @@ struct tcb *tcp;
 		return RVAL_NONE | RVAL_STR;
 	}
 	return 0;
+#else
+#ifdef HPPA
+#warning NO sys_sigreturn DECODE FOR HPPA
+	return 0;
+#endif /* HPPA */
 #endif /* MIPS */
 #endif /* SPARC */
 #endif /* ALPHA */
 #endif /* !M68K */
 #endif /* !POWERPC */
+#endif /* !IA64 */
 #endif /* !I386 */
 #endif /* S390 */
 }
@@ -902,7 +992,7 @@ struct tcb *tcp;
 
 #endif /* LINUX */
 
-#ifdef SVR4
+#if defined(SVR4) || defined(FREEBSD)
 
 int
 sys_sigsuspend(tcp)
@@ -918,6 +1008,7 @@ struct tcb *tcp;
 	}
 	return 0;
 }
+#ifndef FREEBSD
 static struct xlat ucontext_flags[] = {
 	{ UC_SIGMASK,	"UC_SIGMASK"	},
 	{ UC_STACK,	"UC_STACK"	},
@@ -930,10 +1021,10 @@ static struct xlat ucontext_flags[] = {
 #endif
 	{ 0,		NULL		},
 };
+#endif /* !FREEBSD */
+#endif /* SVR4 || FREEBSD */
 
-#endif
-
-#if defined SVR4 || defined LINUX
+#if defined SVR4 || defined LINUX || defined FREEBSD
 #if defined LINUX && !defined SS_ONSTACK
 #define SS_ONSTACK      1
 #define SS_DISABLE      2
@@ -946,6 +1037,9 @@ typedef struct
 } stack_t;
 #endif
 #endif
+#ifdef FREEBSD
+#define stack_t struct sigaltstack
+#endif
 
 static struct xlat sigaltstack_flags[] = {
 	{ SS_ONSTACK,	"SS_ONSTACK"	},
@@ -986,8 +1080,10 @@ struct tcb *tcp;
 {
 	ucontext_t uc;
 
-	if (entering(tcp)) {
-		if (!tcp->u_arg[0])
+	if (exiting(tcp)) {
+		if (tcp->u_error)
+			tprintf("%#lx", tcp->u_arg[0]);
+		else if (!tcp->u_arg[0])
 			tprintf("NULL");
 		else if (umove(tcp, tcp->u_arg[0], &uc) < 0)
 			tprintf("{...}");
@@ -1022,7 +1118,7 @@ struct tcb *tcp;
 
 #endif /* SVR4 */
 
-#ifdef LINUX
+#if defined(LINUX) || defined(FREEBSD)
 
 static int
 print_stack_t(tcp, addr)
