@@ -27,9 +27,10 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: strace.c,v 1.11 1999/12/23 14:20:15 wichert Exp $
+ *	$Id: strace.c,v 1.21 2001/03/28 14:40:14 wichert Exp $
  */
 
+#include <sys/types.h>
 #include "defs.h"
 
 #include <signal.h>
@@ -43,9 +44,12 @@
 #include <grp.h>
 #include <string.h>
 
+#ifdef USE_PROCFS
+#include <poll.h>
+#endif
+
 #ifdef SVR4
 #include <sys/stropts.h>
-#include <poll.h>
 #ifdef HAVE_MP_PROCFS
 #include <sys/uio.h>
 #endif
@@ -86,7 +90,7 @@ static int interrupted;
 #endif /* !__STDC__ */
 #endif /* !HAVE_SIG_ATOMIC_T */
 
-#ifdef SVR4
+#ifdef USE_PROCFS
 
 static struct tcb *pfd2tcb P((int pfd));
 static void reaper P((int sig));
@@ -114,7 +118,7 @@ static int proc_poll_pipe[2] = { -1, -1 };
 #else
 #define POLLWANT	POLLPRI
 #endif
-#endif /* SVR4 */
+#endif /* USE_PROCFS */
 
 static void
 usage(ofp, exitval)
@@ -333,19 +337,19 @@ char *argv[];
 		tcp->outf = outf;
 		if (!(tcp->flags & TCB_INUSE) || !(tcp->flags & TCB_ATTACHED))
 			continue;
-#ifdef SVR4
+#ifdef USE_PROCFS
 		if (proc_open(tcp, 1) < 0) {
 			fprintf(stderr, "trouble opening proc file\n");
 			droptcb(tcp);
 			continue;
 		}
-#else /* !SVR4 */
+#else /* !USE_PROCFS */
 		if (ptrace(PTRACE_ATTACH, tcp->pid, (char *) 1, 0) < 0) {
 			perror("attach: ptrace(PTRACE_ATTACH, ...)");
 			droptcb(tcp);
 			continue;
 		}
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 		if (!qflag)
 			fprintf(stderr,
 				"Process %u attached - interrupt to quit\n",
@@ -407,8 +411,8 @@ char *argv[];
 			exit(1);
 			break;
 		case 0: {
-#ifdef SVR4
-			if (outf != stderr) close (fileno (outf));
+#ifdef USE_PROCFS
+		        if (outf != stderr) close (fileno (outf));
 #ifdef MIPS
 			/* Kludge for SGI, see proc_open for details. */
 			sa.sa_handler = foobar;
@@ -416,8 +420,15 @@ char *argv[];
 			sigemptyset(&sa.sa_mask);
 			sigaction(SIGINT, &sa, NULL);
 #endif /* MIPS */
+#ifndef FREEBSD
 			pause();
-#else /* !SVR4 */
+#else /* FREEBSD */
+			kill(getpid(), SIGSTOP); /* stop HERE */
+#endif /* FREEBSD */			
+#else /* !USE_PROCFS */
+			if (outf!=stderr)	
+				close(fileno (outf));
+
 			if (ptrace(PTRACE_TRACEME, 0, (char *) 1, 0) < 0) {
 				perror("strace: ptrace(PTRACE_TRACEME, ...)");
 				return -1;
@@ -438,23 +449,24 @@ char *argv[];
 				 * It is important to set groups before we
 				 * lose privileges on setuid.
 				 */
-				if (username != NULL
-				    && initgroups(username, run_gid) < 0) {
-					perror("initgroups");
-					exit(1);
-				}
-				if (setregid(run_gid, run_egid) < 0) {
-					perror("setregid");
-					exit(1);
-				}
-				if (setreuid(run_uid, run_euid) < 0) {
-					perror("setreuid");
-					exit(1);
+				if (username != NULL) {
+					if (initgroups(username, run_gid) < 0) {
+						perror("initgroups");
+						exit(1);
+					}
+					if (setregid(run_gid, run_egid) < 0) {
+						perror("setregid");
+						exit(1);
+					}
+					if (setreuid(run_uid, run_euid) < 0) {
+						perror("setreuid");
+						exit(1);
+					}
 				}
 			}
 			else
 				setreuid(run_uid, run_uid);
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 
 			execv(pathname, &argv[optind]);
 			perror("strace: exec");
@@ -467,16 +479,16 @@ char *argv[];
 				cleanup();
 				exit(1);
 			}
-#ifdef SVR4
+#ifdef USE_PROCFS
 			if (proc_open(tcp, 0) < 0) {
 				fprintf(stderr, "trouble opening proc file\n");
 				cleanup();
 				exit(1);
 			}
-#endif /* SVR4 */
-#ifndef SVR4
+#endif /* USE_PROCFS */
+#ifndef USE_PROCFS
 			fake_execve(tcp, pathname, &argv[optind], environ);
-#endif
+#endif /* !USE_PROCFS */
 			break;
 		}
 	}
@@ -507,10 +519,10 @@ char *argv[];
 	sigaction(SIGQUIT, &sa, NULL);
 	sigaction(SIGPIPE, &sa, NULL);
 	sigaction(SIGTERM, &sa, NULL);
-#ifdef SVR4
+#ifdef USE_PROCFS
 	sa.sa_handler = reaper;
 	sigaction(SIGCHLD, &sa, NULL);
-#endif /* SVR4 */
+#endif /* USE_PROCFS */
 
 	if (trace() < 0)
 		exit(1);
@@ -567,7 +579,7 @@ int pid;
 	return NULL;
 }
 
-#ifdef SVR4
+#ifdef USE_PROCFS
 int
 proc_open(tcp, attaching)
 struct tcb *tcp;
@@ -575,9 +587,11 @@ int attaching;
 {
 	char proc[32];
 	long arg;
+#ifdef SVR4
 	sysset_t sc_enter, sc_exit;
 	sigset_t signals;
 	fltset_t faults;
+#endif
 #ifndef HAVE_POLLABLE_PROCFS
 	static int last_pfd;
 #endif
@@ -625,8 +639,13 @@ int attaching;
 	}
 #else
 	/* Open the process pseudo-file in /proc. */
+#ifndef FREEBSD
 	sprintf(proc, "/proc/%d", tcp->pid);
 	if ((tcp->pfd = open(proc, O_RDWR|O_EXCL)) < 0) {
+#else /* FREEBSD */
+	sprintf(proc, "/proc/%d/mem", tcp->pid);
+	if ((tcp->pfd = open(proc, O_RDWR)) < 0) {
+#endif /* FREEBSD */
 		perror("strace: open(\"/proc/...\", ...)");
 		return -1;
 	}
@@ -639,6 +658,21 @@ int attaching;
 		return -1;
 	}
 #endif
+#ifdef FREEBSD
+	sprintf(proc, "/proc/%d/regs", tcp->pid);
+	if ((tcp->pfd_reg = open(proc, O_RDONLY)) < 0) {
+		perror("strace: open(\"/proc/.../regs\", ...)");
+		return -1;
+	}
+	if (cflag) {
+		sprintf(proc, "/proc/%d/status", tcp->pid);
+		if ((tcp->pfd_status = open(proc, O_RDONLY)) < 0) {
+			perror("strace: open(\"/proc/.../status\", ...)");
+			return -1;
+		}
+	} else
+		tcp->pfd_status = -1;
+#endif /* FREEBSD */
 	rebuild_pollv();
 	if (!attaching) {
 		/*
@@ -651,14 +685,16 @@ int attaching;
 				return -1;
 			}
 			if (tcp->status.PR_FLAGS & PR_ASLEEP)
-				break;
+			    break;
 		}
 	}
+#ifndef FREEBSD
 	/* Stop the process so that we own the stop. */
-	if (IOCTL(tcp->pfd, PIOCSTOP, NULL) < 0) {
+	if (IOCTL(tcp->pfd, PIOCSTOP, (char *)NULL) < 0) {
 		perror("strace: PIOCSTOP");
 		return -1;
 	}
+#endif	
 #ifdef PIOCSET
 	/* Set Run-on-Last-Close. */
 	arg = PR_RLC;
@@ -673,6 +709,7 @@ int attaching;
 		return -1;
 	}
 #else  /* !PIOCSET */
+#ifndef FREEBSD	
 	if (ioctl(tcp->pfd, PIOCSRLC) < 0) {
 		perror("PIOCSRLC");
 		return -1;
@@ -681,7 +718,20 @@ int attaching;
 		perror("PIOC{S,R}FORK");
 		return -1;
 	}
+#else /* FREEBSD */
+	/* just unset the PF_LINGER flag for the Run-on-Last-Close. */
+	if (ioctl(tcp->pfd, PIOCGFL, &arg) < 0) {
+	        perror("PIOCGFL");
+	        return -1;
+	}
+	arg &= ~PF_LINGER;
+	if (ioctl(tcp->pfd, PIOCSFL, arg) < 0) {
+	        perror("PIOCSFL");
+	        return -1;
+	}
+#endif /* FREEBSD */
 #endif /* !PIOCSET */
+#ifndef FREEBSD
 	/* Enable all syscall entries. */
 	prfillset(&sc_enter);
 	if (IOCTL(tcp->pfd, PIOCSENTRY, &sc_enter) < 0) {
@@ -706,6 +756,14 @@ int attaching;
 		perror("PIOCSFAULT");
 		return -1;
 	}
+#else /* FREEBSD */
+	/* set events flags. */
+	arg = S_SIG | S_SCE | S_SCX ;
+	if(ioctl(tcp->pfd, PIOCBIS, arg) < 0) {
+		perror("PIOCBIS");
+		return -1;
+	}
+#endif /* FREEBSD */
 	if (!attaching) {
 #ifdef MIPS
 		/*
@@ -714,13 +772,19 @@ int attaching;
 		 */
 		kill(tcp->pid, SIGINT);
 #else /* !MIPS */
+#ifdef PRSABORT	
 		/* The child is in a pause(), abort it. */
 		arg = PRSABORT;
 		if (IOCTL (tcp->pfd, PIOCRUN, &arg) < 0) {
 			perror("PIOCRUN");
 			return -1;
 		}
-#endif /* !MIPS */
+#endif		
+#endif /* !MIPS*/
+#ifdef FREEBSD
+		/* wake up the child if it received the SIGSTOP */
+		kill(tcp->pid, SIGCONT);
+#endif		
 		for (;;) {
 			/* Wait for the child to do something. */
 			if (IOCTL_WSTOP (tcp) < 0) {
@@ -728,22 +792,60 @@ int attaching;
 				return -1;
 			}
 			if (tcp->status.PR_WHY == PR_SYSENTRY) {
-#ifdef HAVE_PR_SYSCALL
-				int scno = tcp->status.pr_syscall;
-#else /* !HAVE_PR_SYSCALL */
-				int scno = tcp->status.PR_WHAT;
-#endif /* !HAVE_PR_SYSCALL */
-				if (scno == SYS_execve)
+				tcp->flags &= ~TCB_INSYSCALL;
+				get_scno(tcp);
+				if (tcp->scno == SYS_execve)
 					break;
 			}
 			/* Set it running: maybe execve will be next. */
+#ifndef FREEBSD
 			arg = 0;
 			if (IOCTL(tcp->pfd, PIOCRUN, &arg) < 0) {
+#else /* FREEBSD */
+			if (IOCTL(tcp->pfd, PIOCRUN, 0) < 0) {
+#endif /* FREEBSD */			  
 				perror("PIOCRUN");
 				return -1;
 			}
+#ifdef FREEBSD
+			/* handle the case where we "opened" the child before
+			   it did the kill -STOP */
+			if (tcp->status.PR_WHY == PR_SIGNALLED &&
+			    tcp->status.PR_WHAT == SIGSTOP)
+			        kill(tcp->pid, SIGCONT);
+#endif			
 		}
+#ifndef FREEBSD
 	}
+#else /* FREEBSD */
+	} else {
+		if (attaching < 2) { 
+			/* We are attaching to an already running process.
+			 * Try to figure out the state of the process in syscalls,
+			 * to handle the first event well.
+			 * This is done by having a look at the "wchan" property of the
+			 * process, which tells where it is stopped (if it is). */
+			FILE * status;
+			char wchan[20]; /* should be enough */
+			
+			sprintf(proc, "/proc/%d/status", tcp->pid);
+			status = fopen(proc, "r");
+			if (status &&
+			    (fscanf(status, "%*s %*d %*d %*d %*d %*d,%*d %*s %*d,%*d"
+				    "%*d,%*d %*d,%*d %19s", wchan) == 1) &&
+			    strcmp(wchan, "nochan") && strcmp(wchan, "spread") &&
+			    strcmp(wchan, "stopevent")) {
+				/* The process is asleep in the middle of a syscall.
+				   Fake the syscall entry event */
+				tcp->flags &= ~(TCB_INSYSCALL|TCB_STARTUP);
+				tcp->status.PR_WHY = PR_SYSENTRY;
+				trace_syscall(tcp);
+			}
+			if (status)
+				fclose(status);
+		} /* otherwise it's a fork being followed */
+	}
+#endif /* FREEBSD */
 #ifndef HAVE_POLLABLE_PROCFS
 	if (proc_poll_pipe[0] != -1)
 		proc_poller(tcp->pfd);
@@ -757,7 +859,7 @@ int attaching;
 	return 0;
 }
 
-#endif /* SVR4 */
+#endif /* USE_PROCFS */
 
 static struct tcb *
 pid2tcb(pid)
@@ -775,7 +877,7 @@ int pid;
 	return NULL;
 }
 
-#ifdef SVR4
+#ifdef USE_PROCFS
 
 static struct tcb *
 pfd2tcb(pfd)
@@ -793,7 +895,7 @@ int pfd;
 	return NULL;
 }
 
-#endif /* SVR4 */
+#endif /* USE_PROCFS */
 
 void
 droptcb(tcp)
@@ -807,7 +909,17 @@ struct tcb *tcp;
 	if (tcp->pfd != -1) {
 		close(tcp->pfd);
 		tcp->pfd = -1;
-#ifdef SVR4
+#ifdef FREEBSD
+		if (tcp->pfd_reg != -1) {
+		        close(tcp->pfd_reg);
+		        tcp->pfd_reg = -1;
+		}
+		if (tcp->pfd_status != -1) {
+			close(tcp->pfd_status);
+			tcp->pfd_status = -1;
+		}
+#endif /* !FREEBSD */		
+#ifdef USE_PROCFS
 		rebuild_pollv();
 #endif
 	}
@@ -822,7 +934,7 @@ struct tcb *tcp;
 	tcp->outf = 0;
 }
 
-#ifndef SVR4
+#ifndef USE_PROCFS
 
 static int
 resume(tcp)
@@ -847,7 +959,7 @@ struct tcb *tcp;
 	return 0;
 }
 
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 
 /* detach traced process; continue with sig */
 
@@ -929,10 +1041,10 @@ int sig;
 		perror("detach: ptrace(PTRACE_DETACH, ...)");
 #endif /* SUNOS4 */
 
-#ifndef SVR4
+#ifndef USE_PROCFS
 	if (waiting_parent(tcp))
 		error = resume(tcp->parent);
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 
 	if (!qflag)
 		fprintf(stderr, "Process %u detached\n", tcp->pid);
@@ -941,7 +1053,7 @@ int sig;
 	return error;
 }
 
-#ifdef SVR4
+#ifdef USE_PROCFS
 
 static void
 reaper(sig)
@@ -961,7 +1073,7 @@ int sig;
 	}
 }
 
-#endif /* SVR4 */
+#endif /* USE_PROCFS */
 
 static void
 cleanup()
@@ -1049,7 +1161,7 @@ int sig;
 
 #endif /* HAVE_STRSIGNAL */
 
-#ifdef SVR4
+#ifdef USE_PROCFS
 
 static void
 rebuild_pollv()
@@ -1136,6 +1248,9 @@ int pfd;
 	int i;
 	int n;
 	struct rlimit rl;
+#ifdef FREEBSD
+	struct procfs_status pfs;
+#endif /* FREEBSD */
 
 	switch (fork()) {
 	case -1:
@@ -1175,7 +1290,11 @@ int pfd;
 	pollinfo.fd = pfd;
 	pollinfo.pid = getpid();
 	for (;;) {
-		if (ioctl(pfd, PIOCWSTOP, NULL) < 0)
+#ifndef FREEBSD
+	        if (ioctl(pfd, PIOCWSTOP, NULL) < 0)
+#else /* FREEBSD */
+	        if (ioctl(pfd, PIOCWSTOP, &pfs) < 0)
+#endif /* FREEBSD */
 		{
 			switch (errno) {
 			case EINTR:
@@ -1326,7 +1445,17 @@ trace()
 	FOUND:
 		/* Get the status of the process. */
 		if (!interrupted) {
+#ifndef FREEBSD
 			ioctl_result = IOCTL_WSTOP (tcp);
+#else /* FREEBSD */
+			/* Thanks to some scheduling mystery, the first poller
+			   sometimes waits for the already processed end of fork
+			   event. Doing a non blocking poll here solves the problem. */
+			if (proc_poll_pipe[0] != -1)
+				ioctl_result = IOCTL_STATUS (tcp);
+			else
+			  	ioctl_result = IOCTL_WSTOP (tcp);
+#endif /* FREEBSD */			  
 			ioctl_errno = errno;
 #ifndef HAVE_POLLABLE_PROCFS
 			if (proc_poll_pipe[0] != -1) {
@@ -1349,6 +1478,9 @@ trace()
 			case EINTR:
 			case EBADF:
 				continue;
+#ifdef FREEBSD
+			case ENOTTY:
+#endif			  
 			case ENOENT:
 				droptcb(tcp);
 				continue;
@@ -1358,6 +1490,14 @@ trace()
 			}
 		}
 
+#ifdef FREEBSD
+		if ((tcp->flags & TCB_STARTUP) && (tcp->status.PR_WHY == PR_SYSEXIT)) {
+			/* discard first event for a syscall we never entered */
+			IOCTL (tcp->pfd, PIOCRUN, 0);
+			continue;
+		}
+#endif		
+		
 		/* clear the just started flag */
 		tcp->flags &= ~TCB_STARTUP;
 
@@ -1366,15 +1506,27 @@ trace()
 
 		if (cflag) {
 			struct timeval stime;
+#ifdef FREEBSD
+			char buf[1024];
+			int len;
 
+			if ((len = pread(tcp->pfd_status, buf, sizeof(buf) - 1, 0)) > 0) {
+				buf[len] = '\0';
+				sscanf(buf,
+				       "%*s %*d %*d %*d %*d %*d,%*d %*s %*d,%*d %*d,%*d %ld,%ld",
+				       &stime.tv_sec, &stime.tv_usec);
+			} else
+				stime.tv_sec = stime.tv_usec = 0;
+#else /* !FREEBSD */			
 			stime.tv_sec = tcp->status.pr_stime.tv_sec;
 			stime.tv_usec = tcp->status.pr_stime.tv_nsec/1000;
+#endif /* !FREEBSD */
 			tv_sub(&tcp->dtime, &stime, &tcp->stime);
 			tcp->stime = stime;
 		}
-
 		what = tcp->status.PR_WHAT;
 		switch (tcp->status.PR_WHY) {
+#ifndef FREEBSD
 		case PR_REQUESTED:
 			if (tcp->status.PR_FLAGS & PR_ASLEEP) {
 				tcp->status.PR_WHY = PR_SYSENTRY;
@@ -1384,6 +1536,7 @@ trace()
 				}
 			}
 			break;
+#endif /* !FREEBSD */
 		case PR_SYSENTRY:
 #ifdef POLL_HACK
 		        in_syscall = tcp;
@@ -1409,13 +1562,21 @@ trace()
 				printtrailer(tcp);
 			}
 			break;
+#ifdef FREEBSD
+		case 0: /* handle case we polled for nothing */
+		  	continue;
+#endif			
 		default:
 			fprintf(stderr, "odd stop %d\n", tcp->status.PR_WHY);
 			exit(1);
 			break;
 		}
 		arg = 0;
+#ifndef FREEBSD		
 		if (IOCTL (tcp->pfd, PIOCRUN, &arg) < 0) {
+#else		  
+		if (IOCTL (tcp->pfd, PIOCRUN, 0) < 0) {
+#endif		  
 			perror("PIOCRUN");
 			exit(1);
 		}
@@ -1423,7 +1584,7 @@ trace()
 	return 0;
 }
 
-#else /* !SVR4 */
+#else /* !USE_PROCFS */
 
 static int
 trace()
@@ -1434,13 +1595,36 @@ trace()
 	struct tcb *tcp;
 #ifdef LINUX
 	struct rusage ru;
+#ifdef __WALL
+	static int wait4_options = __WALL;
+#endif
 #endif /* LINUX */
 
 	while (nprocs != 0) {
 		if (interactive)
 			sigprocmask(SIG_SETMASK, &empty_set, NULL);
 #ifdef LINUX
+#ifdef __WALL
+		pid = wait4(-1, &status, wait4_options, cflag ? &ru : NULL);
+		if ((wait4_options & __WALL) && errno == EINVAL) {
+			/* this kernel does not support __WALL */
+			wait4_options &= ~__WALL;
+			errno = 0;
+			pid = wait4(-1, &status, wait4_options,
+					cflag ? &ru : NULL);
+		}
+		if (!(wait4_options & __WALL) && errno == ECHILD) {
+			/* most likely a "cloned" process */
+			pid = wait4(-1, &status, __WCLONE,
+					cflag ? &ru : NULL);
+			if (pid == -1) {
+				fprintf(stderr, "strace: clone wait4 "
+						"failed: %s\n", strerror(errno));
+			}
+		}
+#else
 		pid = wait4(-1, &status, 0, cflag ? &ru : NULL);
+#endif /* __WALL */
 #endif /* LINUX */
 #ifdef SUNOS4
 		pid = wait(&status);
@@ -1481,10 +1665,25 @@ trace()
 
 		/* Look up `pid' in our table. */
 		if ((tcp = pid2tcb(pid)) == NULL) {
+#if 0 /* XXX davidm */ /* WTA: disabled again */
+			struct tcb *tcpchild;
+
+			if ((tcpchild = alloctcb(pid)) == NULL) {
+				fprintf(stderr, " [tcb table full]\n");
+				kill(pid, SIGKILL); /* XXX */
+				return 0;
+			}
+			tcpchild->flags |= TCB_ATTACHED;
+			newoutf(tcpchild);
+			tcp->nchildren++;
+			if (!qflag)
+				fprintf(stderr, "Process %d attached\n", pid);
+#else
 			fprintf(stderr, "unknown pid: %u\n", pid);
 			if (WIFSTOPPED(status))
 				ptrace(PTRACE_CONT, pid, (char *) 1, 0);
 			exit(1);
+#endif
 		}
 		/* set current output file */
 		outf = tcp->outf;
@@ -1647,7 +1846,7 @@ trace()
 	return 0;
 }
 
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 
 static int curcol;
 
