@@ -25,15 +25,17 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: stream.c,v 1.5 1999/12/23 14:20:15 wichert Exp $
+ *	$Id: stream.c,v 1.9 2000/10/13 12:47:12 wichert Exp $
  */
 
 #include "defs.h"
 
-#if defined(HAVE_SYS_STREAM_H) || defined(linux)
+#if defined(HAVE_SYS_STREAM_H) || defined(linux) || defined(FREEBSD)
 
 #if defined(linux)
+#ifdef HAVE_SYS_POLL_H
 #include <sys/poll.h>
+#endif
 
 #define RS_HIPRI 1
 struct strbuf {
@@ -46,11 +48,15 @@ struct strbuf {
 
 #else /* linux */
 
+#ifndef FREEBSD
 #include <stropts.h>
 #include <poll.h>
 #include <sys/conf.h>
 #include <sys/stream.h>
 #include <sys/tihdr.h>
+#else /* FREEBSD */
+#include <poll.h>
+#endif /* FREEBSD */
 
 #endif /* linux */
 
@@ -60,6 +66,7 @@ struct strbuf {
 #include <sys/timod.h>
 #endif /* HAVE_SYS_TIUSER_H */
 
+#ifndef FREEBSD
 static struct xlat msgflags[] = {
 	{ RS_HIPRI,	"RS_HIPRI"	},
 	{ 0,		NULL		},
@@ -171,9 +178,15 @@ struct tcb *tcp;
 #ifdef HAVE_PUTPMSG
 
 static struct xlat pmsgflags[] = {
+#ifdef MSG_HIPRI
 	{ MSG_HIPRI,	"MSG_HIPRI"	},
+#endif
+#ifdef MSG_AND
 	{ MSG_ANY,	"MSG_ANY"	},
+#endif
+#ifdef MSG_BAND
 	{ MSG_BAND,	"MSG_BAND"	},
+#endif
 	{ 0,		NULL		},
 };
 
@@ -250,9 +263,13 @@ struct tcb *tcp;
 }
 
 #endif /* HAVE_PUTPMSG */
+#endif /* !FREEBSD */
 
 
+#ifdef HAVE_SYS_POLL_H
+
 static struct xlat pollflags[] = {
+#ifdef POLLIN
 	{ POLLIN,	"POLLIN"	},
 	{ POLLPRI,	"POLLPRI"	},
 	{ POLLOUT,	"POLLOUT"	},
@@ -271,6 +288,7 @@ static struct xlat pollflags[] = {
 	{ POLLERR,	"POLLERR"	},
 	{ POLLHUP,	"POLLHUP"	},
 	{ POLLNVAL,	"POLLNVAL"	},
+#endif
 	{ 0,		NULL		},
 };
 
@@ -334,7 +352,16 @@ struct tcb *tcp;
 	return 0;
 }
 
-#ifndef linux
+#else /* !HAVE_SYS_POLL_H */
+int
+sys_poll(tcp)
+struct tcb *tcp;
+{
+    	return 0;
+}
+#endif
+
+#if !defined(linux) && !defined(FREEBSD)
 
 static struct xlat stream_flush_options[] = {
 	{ FLUSHR,	"FLUSHR"	},
@@ -809,7 +836,7 @@ int code, arg;
 	}
 }
 
-#endif /* linux */
+#endif /* !linux && !FREEBSD */ 
 
-#endif /* LINUXSPARC && linux */
+#endif /* HAVE_SYS_STREAM_H || linux || FREEBSD */
 
