@@ -24,7 +24,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: dummy.h,v 1.2 1999/06/11 13:18:41 wichert Exp $
+ *	$Id: dummy.h,v 1.7 2001/03/07 14:47:19 hughesj Exp $
  */
 
 /* still unfinished */
@@ -58,6 +58,7 @@
 #define sys_priocntl printargs
 #define sys_ksigqueue printargs
 #define sys_lwp_sema_wait printargs
+#define sys_lwp_sema_trywait printargs
 #define sys_memcntl printargs
 #define sys_syscall printargs
 #define sys_clocal printargs
@@ -78,7 +79,6 @@
 #define sys_fstatfs printargs
 #define sys_statvfs printargs
 #define sys_fstatvfs printargs
-#define sys_fork1 printargs
 #define sys_sigsendsys printargs
 #define sys_gtty printargs
 #define sys_vtrace printargs
@@ -109,7 +109,6 @@
 #define sys_lwp_cond_wait printargs
 #define sys_lwp_cond_signal printargs
 #define sys_lwp_cond_broadcast printargs
-#define sys_llseek printargs
 #define sys_inst_sync printargs
 #define sys_auditsys printargs
 #define sys_processor_bind printargs
@@ -125,12 +124,6 @@
 #define sys_timer_settime printargs
 #define sys_timer_gettime printargs
 #define sys_timer_getoverrun printargs
-#define sys_signal printargs
-#define sys_sigset printargs
-#define sys_sighold printargs
-#define sys_sigrelse printargs
-#define sys_sigignore printargs
-#define sys_sigpause printargs
 #define sys_msgctl printargs
 #define sys_msgget printargs
 #define sys_msgrcv printargs
@@ -148,6 +141,76 @@
 #define sys_sysfs1 printargs
 #define sys_sysfs2 printargs
 #define sys_sysfs3 printargs
+#define sys_keyctl printargs
+#define sys_secsys printargs
+#define sys_filepriv printargs
+#define sys_procpriv printargs
+#define sys_devstat printargs
+#define sys_fdevstat printargs
+#define sys_flvlfile printargs
+#define sys_lvlfile printargs
+#define sys_lvlequal printargs
+#define sys_lvlproc printargs
+#define sys_lvlipc printargs
+#define sys_auditevt printargs
+#define sys_auditctl printargs
+#define sys_auditdmp printargs
+#define sys_auditlog printargs
+#define sys_auditbuf printargs
+#define sys_lvldom printargs
+#define sys_lvlvfs printargs
+#define sys_mkmld printargs
+#define sys_mldmode printargs
+#define sys_secadvise printargs
+#define sys_online printargs
+#define sys_lwpinfo printargs
+#define sys_lwpprivate printargs
+#define sys_processor_exbind printargs
+#define sys_prepblock printargs
+#define sys_block printargs
+#define sys_rdblock printargs
+#define sys_unblock printargs
+#define sys_cancelblock printargs
+#define sys_lwpkill printargs
+#define sys_sigwait printargs
+#define sys_modload printargs
+#define sys_moduload printargs
+#define sys_modpath printargs
+#define sys_modstat printargs
+#define sys_modadm printargs
+#define sys_lwpsuspend printargs
+#define sys_lwpcontinue printargs
+#define sys_priocntllst printargs
+#define sys_lwp_sema_trywait printargs
+#define sys_xsetsockaddr printargs
+#define sys_dshmsys printargs
+#define sys_invlpg printargs
+#define sys_rfork1 printargs
+#define sys_rforkall printargs
+#define sys_rexecve printargs
+#define sys_migrate printargs
+#define sys_kill3 printargs
+#define sys_ssisys printargs
+#define sys_xbindresvport printargs
+#define sys_lwp_sema_trywait printargs
+#define sys_tsolsys printargs
+#ifndef HAVE_SYS_ACL_H
+#define sys_acl printargs
+#define sys_facl printargs
+#define sys_aclipc printargs
+#endif
+#define sys_install_utrap printargs
+#define sys_signotify printargs
+#define sys_schedctl printargs
+#define sys_pset printargs
+#define sys_resolvepath printargs
+#define sys_signotifywait printargs
+#define sys_lwp_sigredirect printargs
+#define sys_lwp_alarm printargs
+#define sys_rpcsys printargs
+#define sys_sockconfig printargs
+#define sys_ntp_gettime printargs
+#define sys_ntp_adjtime printargs
 
 /* like another call */
 #define sys_lchown sys_chown
@@ -165,6 +228,59 @@
 #define sys_BSDgetpgrp sys_getpgrp
 #define sys_BSDsetpgrp sys_setpgrp
 #define sys_waitsys sys_waitid
+#define sys_sigset sys_signal
+#define sys_sigrelse sys_sighold
+#define sys_sigignore sys_sighold
+#define sys_sigpause sys_sighold
+#define sys_sleep sys_alarm
+#define sys_fork1 sys_fork
+#define sys_forkall sys_fork
+
+/* aio */
+#define sys_aionotify printargs
+#define sys_aioinit printargs
+#define sys_aiostart printargs
+#define sys_aiolio printargs
+#define sys_aiosuspend printargs
+#define sys_aioerror printargs
+#define sys_aioliowait printargs
+#define sys_aioaread printargs
+#define sys_aioawrite printargs
+#define sys_aiolio64 printargs
+#define sys_aiosuspend64 printargs
+#define sys_aioerror64 printargs
+#define sys_aioliowait64 printargs
+#define sys_aioaread64 printargs
+#define sys_aioaread64 printargs
+#define sys_aioawrite64 printargs
+#define sys_aiocancel64 printargs
+#define sys_aiofsync printargs
+
+/* the various 64-bit file stuff */
+#if !_LFS64_LARGEFILE
+/* we've implemented these */
+#define sys_getdents64 printargs
+#define sys_mmap64 printargs
+#define sys_stat64 printargs
+#define sys_lstat64 printargs
+#define sys_fstat64 printargs
+#define sys_setrlimit64 printargs
+#define sys_getrlimit64 printargs
+#define sys_pread64 printargs
+#define sys_pwrite64 printargs
+#define sys_ftruncate64 printargs
+#define sys_truncate64 printargs
+#define sys_lseek64 printargs
+#endif
+
+/* unimplemented 64-bit stuff */
+#define sys_statvfs64 printargs
+#define sys_fstatvfs64 printargs
+
+/* like another call */
+#define sys_creat64 sys_creat
+#define sys_open64 sys_open
+#define sys_llseek sys_lseek64
 
 /* printargs does the right thing */
 #define sys_sync printargs
@@ -184,6 +300,7 @@
 #define sys_spcall printargs
 #define sys_context printargs
 #define sys_door printargs
+#define sys_kaio printargs
 
 #if DONE
 #define sys_mount printargs
@@ -276,4 +393,5 @@
 #define sys_gettimeofday printargs
 #define sys_getitimer printargs
 #define sys_setitimer printargs
+#define sys_settimeofday printargs
 #endif
