@@ -30,7 +30,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: syscall.c,v 1.15 1999/12/23 15:08:17 wichert Exp $
+ *	$Id: syscall.c,v 1.30 2001/03/28 20:29:17 wichert Exp $
  */
 
 #include "defs.h"
@@ -43,19 +43,32 @@
 #include <sys/param.h>
 
 #if HAVE_ASM_REG_H
+#ifdef SPARC
+#  define fpq kernel_fpq
+#  define fq kernel_fq
+#  define fpu kernel_fpu
+#endif
 #include <asm/reg.h>
+#ifdef SPARC
+#  undef fpq
+#  undef fq
+#  undef fpu 
+#endif
 #endif
-
-#if HAVE_LINUX_PTRACE_H
-#undef PTRACE_SYSCALL
-#include <linux/ptrace.h>
-#endif 
 
 #ifdef HAVE_SYS_REG_H
 #include <sys/reg.h>
 #ifndef PTRACE_PEEKUSR
 # define PTRACE_PEEKUSR PTRACE_PEEKUSER
 #endif
+#elif defined(HAVE_LINUX_PTRACE_H)
+#undef PTRACE_SYSCALL
+#include <linux/ptrace.h>
+#endif
+
+#if defined(LINUX) && defined(IA64)
+# include <asm/ptrace_offsets.h>
+# include <asm/rse.h>
 #endif
 
 #ifndef SYS_ERRLIST_DECLARED
@@ -63,6 +76,7 @@ extern int sys_nerr;
 extern char *sys_errlist[];
 #endif /* SYS_ERRLIST_DECLARED */
 
+#define NR_SYSCALL_BASE 0
 #ifdef LINUX
 #ifndef ERESTARTSYS
 #define ERESTARTSYS	512
@@ -82,6 +96,8 @@ extern char *sys_errlist[];
 #ifdef ARM
 #undef NSIG
 #define NSIG 32
+#undef NR_SYSCALL_BASE
+#define NR_SYSCALL_BASE __NR_SYSCALL_BASE
 #endif
 #endif /* LINUX */
 
@@ -257,12 +273,12 @@ char *s;
 	int i;
 	char buf[32];
 
-	if (s && *s && isdigit(*s))
+	if (s && *s && isdigit((unsigned char)*s))
 		return atoi(s);
 	strcpy(buf, s);
 	s = buf;
 	for (i = 0; s[i]; i++)
-		s[i] = toupper(s[i]);
+		s[i] = toupper((unsigned char)(s[i]));
 	if (strncmp(s, "SIG", 3) == 0)
 		s += 3;
 	for (i = 0; i <= NSIG; i++) {
@@ -283,7 +299,7 @@ static int
 lookup_desc(s)
 char *s;
 {
-	if (s && *s && isdigit(*s))
+	if (s && *s && isdigit((unsigned char)*s))
 		return atoi(s);
 	return -1;
 }
@@ -380,11 +396,7 @@ struct tcb *tcp;
 		return;
 	if (tcp->u_arg[0] < 0 || tcp->u_arg[0] >= MAX_QUALS)
 		return;
-#ifdef __arm__
-	switch (tcp->scno + __NR_SYSCALL_BASE) {
-#else
-	switch (tcp->scno) {
-#endif
+	switch (tcp->scno + NR_SYSCALL_BASE) {
 	case SYS_read:
 #ifdef SYS_recv
 	case SYS_recv:
@@ -408,9 +420,29 @@ struct tcb *tcp;
 	}
 }
 
+#ifndef FREEBSD
 enum subcall_style { shift_style, deref_style, mask_style, door_style };
+#else /* FREEBSD */
+enum subcall_style { shift_style, deref_style, mask_style, door_style, table_style };
 
-#if !(defined(LINUX) && ( defined(ALPHA) || defined(MIPS) ))
+struct subcall {
+  int call;
+  int nsubcalls;
+  int subcalls[5];
+};
+
+const struct subcall subcalls_table[] = {
+  { SYS_shmsys, 5, { SYS_shmat, SYS_shmctl, SYS_shmdt, SYS_shmget, SYS_shmctl } },
+#ifdef SYS_semconfig
+  { SYS_semsys, 4, { SYS___semctl, SYS_semget, SYS_semop, SYS_semconfig } },
+#else
+  { SYS_semsys, 3, { SYS___semctl, SYS_semget, SYS_semop } },
+#endif
+  { SYS_msgsys, 4, { SYS_msgctl, SYS_msgget, SYS_msgsnd, SYS_msgrcv } },
+};
+#endif /* FREEBSD */
+
+#if !(defined(LINUX) && ( defined(ALPHA) || defined(IA64) || defined(MIPS) ))
 
 const int socket_map [] = {
 	       /* SYS_SOCKET      */ 97,
@@ -463,10 +495,10 @@ enum subcall_style style;
 {
 	int i, addr, mask, arg;
 
-	if (tcp->u_arg[0] < 0 || tcp->u_arg[0] >= nsubcalls)
-		return;
 	switch (style) {
 	case shift_style:
+		if (tcp->u_arg[0] < 0 || tcp->u_arg[0] >= nsubcalls)
+			return;
 		tcp->scno = subcall + tcp->u_arg[0];
 		if (sysent[tcp->scno].nargs != -1)
 			tcp->u_nargs = sysent[tcp->scno].nargs;
@@ -476,6 +508,8 @@ enum subcall_style style;
 			tcp->u_arg[i] = tcp->u_arg[i + 1];
 		break;
 	case deref_style:
+		if (tcp->u_arg[0] < 0 || tcp->u_arg[0] >= nsubcalls)
+			return;
 		tcp->scno = subcall + tcp->u_arg[0];
 		addr = tcp->u_arg[1];
 		for (i = 0; i < sysent[tcp->scno].nargs; i++) {
@@ -488,9 +522,11 @@ enum subcall_style style;
 		break;
 	case mask_style:
 		mask = (tcp->u_arg[0] >> 8) & 0xff;
-		tcp->u_arg[0] &= 0xff;
 		for (i = 0; mask; i++)
 			mask >>= 1;
+		if (i >= nsubcalls)
+			return;
+		tcp->u_arg[0] &= 0xff;
 		tcp->scno = subcall + i;
 		if (sysent[tcp->scno].nargs != -1)
 			tcp->u_nargs = sysent[tcp->scno].nargs;
@@ -498,13 +534,28 @@ enum subcall_style style;
 	case door_style:
 		/*
 		 * Oh, yuck.  The call code is the *sixth* argument.
+		 * (don't you mean the *last* argument? - JH)
 		 */
+		if (tcp->u_arg[5] < 0 || tcp->u_arg[5] >= nsubcalls)
+			return;
 		tcp->scno = subcall + tcp->u_arg[5];
 		if (sysent[tcp->scno].nargs != -1)
 			tcp->u_nargs = sysent[tcp->scno].nargs;
 		else
 			tcp->u_nargs--;
 		break;
+#ifdef FREEBSD
+	case table_style:
+		for (i = 0; i < sizeof(subcalls_table) / sizeof(struct subcall); i++)
+			if (subcalls_table[i].call == tcp->scno) break;
+		if (i < sizeof(subcalls_table) / sizeof(struct subcall) &&
+		    tcp->u_arg[0] >= 0 && tcp->u_arg[0] < subcalls_table[i].nsubcalls) {
+			tcp->scno = subcalls_table[i].subcalls[tcp->u_arg[0]];
+			for (i = 0; i < tcp->u_nargs; i++)
+				tcp->u_arg[i] = tcp->u_arg[i + 1];
+		}
+		break;
+#endif /* FREEBSD */
 	}
 }
 #endif
@@ -520,11 +571,7 @@ struct tcb *tcp;
 	 * correctly support following forks in the presence of tracing
 	 * qualifiers.
 	 */
-#ifdef __arm__
-	switch (tcp->scno + __NR_SYSCALL_BASE) {
-#else
-	switch (tcp->scno) {
-#endif
+	switch (tcp->scno + NR_SYSCALL_BASE) {
 #ifdef SYS_fork
 	case SYS_fork:
 #endif
@@ -571,43 +618,48 @@ struct tcb *tcp;
 	return 0;
 }
 
-int
-trace_syscall(tcp)
-struct tcb *tcp;
-{
-	int sys_res;
-	struct timeval tv;
-	long scno = 0;
+
 #ifdef LINUX
 #if defined (I386)
-	long eax;
+	static long eax;
+#elif defined (IA64)
+	long r8, r10, psr;
+	long ia32 = 0;
 #elif defined (POWERPC)
-	long result,flags;
+	static long result,flags;
 #elif defined (M68K)
-	int d0;
+	static int d0;
 #elif defined (ARM)
-	int r0;
+	static int r0;
 #elif defined (ALPHA)
-	long r0;
-	long a3;
-#elif defined(MIPS)
-	long r2,a3;
+	static long r0;
+	static long a3;
 #elif defined (SPARC)
-	struct regs regs;
-	unsigned long trap;
+	static struct regs regs;
+	static unsigned long trap;
+#elif defined(MIPS)
+	static long a3;
+	static long r2;
 #elif defined(S390)
-	long gpr2;
-	long pc;
+	static long gpr2;
+	static long pc;
+#elif defined(HPPA)
+	static long r28;
 #endif 
 #endif /* LINUX */
+#ifdef FREEBSD
+	struct reg regs;
+#endif /* FREEBSD */	
 
-#ifndef SVR4
+int
+get_scno(tcp)
+struct tcb *tcp;
+{
+	long scno = 0;
+#ifndef USE_PROCFS
 	int pid = tcp->pid;
-#endif /* !SVR4 */
+#endif /* !PROCFS */	
 
-	/* Measure the exit time as early as possible to avoid errors. */
-	if (dtime && (tcp->flags & TCB_INSYSCALL))
-		gettimeofday(&tv, NULL);
 #ifdef LINUX
 #if defined(S390)
 	if (upeek(tcp->pid,PT_PSWADDR,&pc) < 0)
@@ -629,6 +681,25 @@ struct tcb *tcp;
 #elif defined (I386)
 	if (upeek(pid, 4*ORIG_EAX, &scno) < 0)
 		return -1;
+#elif defined(IA64)
+#define IA64_PSR_IS	((long)1 << 34)
+	if (upeek (pid, PT_CR_IPSR, &psr) >= 0)
+		ia32 = (psr & IA64_PSR_IS);
+	if (!(tcp->flags & TCB_INSYSCALL)) {
+		if (ia32) {
+			if (upeek(pid, PT_R8, &scno) < 0)
+				return -1;
+		} else {
+			if (upeek (pid, PT_R15, &scno) < 0)
+				return -1;
+		}
+	} else {
+		/* syscall in progress */
+		if (upeek (pid, PT_R8, &r8) < 0)
+			return -1;
+		if (upeek (pid, PT_R10, &r10) < 0)
+			return -1;
+	}
 #elif defined (ARM)
 	{ 
 	    long pc;
@@ -747,18 +818,60 @@ struct tcb *tcp;
 			memmove (&regs.r_o0, &regs.r_o1, 7*sizeof(regs.r_o0));
 		}
 	}
+#elif defined(HPPA)
+	if (upeek(pid, PT_GR20, &scno) < 0)
+		return -1;
+	if (!(tcp->flags & TCB_INSYSCALL)) {
+		/* Check if we return from execve. */
+		if ((tcp->flags & TCB_WAITEXECVE)) {
+			tcp->flags &= ~TCB_WAITEXECVE;
+			return 0;
+		}
+	}
 #endif 
 #endif /* LINUX */
 #ifdef SUNOS4
 	if (upeek(pid, uoff(u_arg[7]), &scno) < 0)
 		return -1;
 #endif
-#ifdef SVR4
+#ifdef USE_PROCFS
 #ifdef HAVE_PR_SYSCALL
-	scno = tcp->status.pr_syscall;
+	scno = tcp->status.PR_SYSCALL;
 #else /* !HAVE_PR_SYSCALL */
+#ifndef FREEBSD
 	scno = tcp->status.PR_WHAT;
+#else /* FREEBSD */
+	if (pread(tcp->pfd_reg, &regs, sizeof(regs), 0) < 0) {
+	        perror("pread");
+                return -1;
+        }
+	switch (regs.r_eax) {
+	case SYS_syscall:
+	case SYS___syscall:
+    	        pread(tcp->pfd, &scno, sizeof(scno), regs.r_esp + sizeof(int));
+	        break;
+	default:
+	        scno = regs.r_eax;
+	        break;
+	}
+#endif /* FREEBSD */
 #endif /* !HAVE_PR_SYSCALL */
+#endif /* USE_PROCFS */
+	if (!(tcp->flags & TCB_INSYSCALL))
+		tcp->scno = scno;
+	return 1;
+}
+
+
+int
+syscall_fixup(tcp)
+struct tcb *tcp;
+{
+#ifndef USE_PROCFS
+	int pid = tcp->pid;
+#else /* USE_PROCFS */	
+	int scno = tcp->scno;
+
 	if (!(tcp->flags & TCB_INSYSCALL)) {
 		if (tcp->status.PR_WHY != PR_SYSENTRY) {
 			if (
@@ -784,7 +897,7 @@ struct tcb *tcp;
 			tcp->flags &= ~TCB_INSYSCALL;
 		}
 	}
-#endif /* SVR4 */
+#endif /* USE_PROCFS */
 #ifdef SUNOS4
 	if (!(tcp->flags & TCB_INSYSCALL)) {
 		if (scno == 0) {
@@ -815,6 +928,14 @@ struct tcb *tcp;
 			fprintf(stderr, "stray syscall exit: eax = %ld\n", eax);
 		return 0;
 	}
+#elif defined (S390)
+	if (upeek(pid, PT_GPR2, &gpr2) < 0)
+		return -1;
+	if (gpr2 != -ENOSYS && !(tcp->flags & TCB_INSYSCALL)) {
+		if (debug)
+			fprintf(stderr, "stray syscall exit: gpr2 = %ld\n", gpr2);
+		return 0;
+	}
 #elif defined (POWERPC)
 # define SO_MASK 0x10000000
 	if (upeek(pid, 4*PT_CCR, &flags) < 0)
@@ -839,14 +960,30 @@ struct tcb *tcp;
 			fprintf(stderr, "stray syscall exit: d0 = %ld\n", r0);
 		return 0;
 	}
-#else
+#elif defined (HPPA)
+	if (upeek(pid, PT_GR28, &r28) < 0)
+		return -1;
 #endif
 #endif /* LINUX */
+	return 1;
+}
 
-	if (tcp->flags & TCB_INSYSCALL) {
-		long u_error;
-
+int
+get_error(tcp)
+struct tcb *tcp;
+{
+	int u_error = 0;
 #ifdef LINUX
+#ifdef S390
+		if (gpr2 && (unsigned) -gpr2 < nerrnos) {
+			tcp->u_rval = -1;
+			u_error = -gpr2;
+		}
+		else {
+			tcp->u_rval = gpr2;
+			u_error = 0;
+		}
+#else /* !S390 */
 #ifdef I386
 		if (eax < 0 && -eax < nerrnos) {
 			tcp->u_rval = -1;
@@ -857,6 +994,29 @@ struct tcb *tcp;
 			u_error = 0;
 		}
 #else /* !I386 */
+#ifdef IA64
+		if (ia32) {
+			int err;
+
+			err = (int)r8;
+			if (err < 0 && -err < nerrnos) {
+				tcp->u_rval = -1;
+				u_error = -err;
+			}
+			else {
+				tcp->u_rval = err;
+				u_error = 0;
+			}
+		} else {
+			if (r10) {
+				tcp->u_rval = -1;
+				u_error = r8;
+			} else {
+				tcp->u_rval = r8;
+				u_error = 0;
+			}
+		}
+#else /* !IA64 */
 #ifdef MIPS
 		if (a3) {
 		  	tcp->u_rval = -1;
@@ -915,13 +1075,26 @@ struct tcb *tcp;
 			tcp->u_rval = regs.r_o0;
 			u_error = 0;
 		}
+#else /* !SPARC */
+#ifdef HPPA
+		if (r28 && (unsigned) -r28 < nerrnos) {
+			tcp->u_rval = -1;
+			u_error = -r28;
+		}
+		else {
+			tcp->u_rval = r28;
+			u_error = 0;
+		}
+#endif /* HPPA */
 #endif /* SPARC */
 #endif /* ALPHA */
 #endif /* ARM */
 #endif /* M68K */
 #endif /* POWERPC */
 #endif /* MIPS */
+#endif /* IA64 */
 #endif /* I386 */
+#endif /* S390 */
 #endif /* LINUX */
 #ifdef SUNOS4
 		/* get error code from user struct */
@@ -953,6 +1126,11 @@ struct tcb *tcp;
 		}
 		else {
 			tcp->u_rval = tcp->status.PR_REG[EAX];
+#ifdef HAVE_LONG_LONG
+			tcp->u_lrval =
+				((unsigned long long) tcp->status.PR_REG[EDX] << 32) +
+				tcp->status.PR_REG[EAX];
+#endif
 			u_error = 0;
 		}
 #endif /* I386 */
@@ -967,10 +1145,278 @@ struct tcb *tcp;
 		}
 #endif /* MIPS */
 #endif /* SVR4 */
-		tcp->u_error = u_error;
+#ifdef FREEBSD
+		if (regs.r_eflags & PSL_C) {
+ 		        tcp->u_rval = -1;
+		        u_error = regs.r_eax;
+		} else {
+		        tcp->u_rval = regs.r_eax;
+			tcp->u_lrval =
+			  ((unsigned long long) regs.r_edx << 32) +  regs.r_eax;
+		        u_error = 0;
+		}
+#endif /* FREEBSD */	
+	tcp->u_error = u_error;
+	return 1;
+}
+
+int syscall_enter(tcp)
+struct tcb *tcp;
+{
+#ifndef USE_PROCFS
+	int pid = tcp->pid;
+#endif /* !USE_PROCFS */	
+#ifdef LINUX
+#if defined(S390)
+	{
+		int i;
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	tcp->u_nargs = MAX_ARGS;
+		for (i = 0; i < tcp->u_nargs; i++) {
+			if (upeek(pid,i==0 ? PT_ORIGGPR2:PT_GPR2+(i<<2), &tcp->u_arg[i]) < 0)
+				return -1;
+		}
+	}
+#elif defined (ALPHA)
+	{
+		int i;
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	tcp->u_nargs = MAX_ARGS;
+		for (i = 0; i < tcp->u_nargs; i++) {
+			/* WTA: if scno is out-of-bounds this will bomb. Add range-check
+			 * for scno somewhere above here!
+			 */
+			if (upeek(pid, REG_A0+i, &tcp->u_arg[i]) < 0)
+				return -1;
+		}
+	}
+#elif defined (IA64)
+	{
+		unsigned long *bsp, cfm, i;
+
+		if (upeek(pid, PT_AR_BSP, (long *) &bsp) < 0)
+			return -1;
+		if (upeek(pid, PT_CFM, (long *) &cfm) < 0)
+			return -1;
+
+		bsp = ia64_rse_skip_regs(bsp, -(cfm & 0x7f));
+
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	tcp->u_nargs = MAX_ARGS;
+		for (i = 0; i < tcp->u_nargs; ++i) {
+			if (umoven(tcp, (unsigned long) ia64_rse_skip_regs(bsp, i), sizeof(long),
+				   (char *) &tcp->u_arg[i])
+			    < 0)
+				return -1;
+		}
+	}
+#elif defined (MIPS)
+	{
+	  	long sp;
+	  	int i, nargs;
+
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			nargs = tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	nargs = tcp->u_nargs = MAX_ARGS;
+		if(nargs > 4) {
+		  	if(upeek(pid, REG_SP, &sp) < 0)
+			  	return -1;
+			for(i = 0; i < 4; i++) {
+			  	if (upeek(pid, REG_A0 + i, &tcp->u_arg[i])<0)
+				  	return -1;
+			}
+			umoven(tcp, sp+16, (nargs-4) * sizeof(tcp->u_arg[0]),
+			       (char *)(tcp->u_arg + 4));
+		} else {
+		  	for(i = 0; i < nargs; i++) {
+			  	if (upeek(pid, REG_A0 + i, &tcp->u_arg[i]) < 0)
+				  	return -1;
+			}
+		}
+	}
+#elif defined (POWERPC)
+	{
+		int i;
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	tcp->u_nargs = MAX_ARGS;
+		for (i = 0; i < tcp->u_nargs; i++) {
+			if (upeek(pid, (i==0) ? (4*PT_ORIG_R3) : ((i+PT_R3)*4), &tcp->u_arg[i]) < 0)
+				return -1;
+		}
+	}
+#elif defined (SPARC)
+	{
+		int i;
+
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	tcp->u_nargs = MAX_ARGS;
+		for (i = 0; i < tcp->u_nargs; i++)
+			tcp->u_arg[i] = *((&regs.r_o0) + i);
+	}
+#elif defined (HPPA)
+	{
+		int i;
+
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	tcp->u_nargs = MAX_ARGS;
+		for (i = 0; i < tcp->u_nargs; i++) {
+			if (upeek(pid, PT_GR26-4*i, &tcp->u_arg[i]) < 0)
+				return -1;
+		}
+	}
+#else /* Other architecture (like i386) (32bits specific) */
+	{
+		int i;
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	tcp->u_nargs = MAX_ARGS;
+		for (i = 0; i < tcp->u_nargs; i++) {
+			if (upeek(pid, i*4, &tcp->u_arg[i]) < 0)
+				return -1;
+		}
+	}
+#endif 
+#endif /* LINUX */
+#ifdef SUNOS4
+	{
+		int i;
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+			tcp->u_nargs = sysent[tcp->scno].nargs;
+		else 
+     	        	tcp->u_nargs = MAX_ARGS;
+		for (i = 0; i < tcp->u_nargs; i++) {
+			struct user *u;
+
+			if (upeek(pid, uoff(u_arg[0]) +
+			    (i*sizeof(u->u_arg[0])), &tcp->u_arg[i]) < 0)
+				return -1;
+		}
+	}
+#endif /* SUNOS4 */
+#ifdef SVR4
+#ifdef MIPS
+	/*
+	 * SGI is broken: even though it has pr_sysarg, it doesn't
+	 * set them on system call entry.  Get a clue.
+	 */
+	if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+		tcp->u_nargs = sysent[tcp->scno].nargs;
+	else
+		tcp->u_nargs = tcp->status.pr_nsysarg;
+	if (tcp->u_nargs > 4) {
+		memcpy(tcp->u_arg, &tcp->status.pr_reg[CTX_A0],
+			4*sizeof(tcp->u_arg[0]));
+		umoven(tcp, tcp->status.pr_reg[CTX_SP] + 16,
+			(tcp->u_nargs - 4)*sizeof(tcp->u_arg[0]), (char *) (tcp->u_arg + 4));
+	}
+	else {
+		memcpy(tcp->u_arg, &tcp->status.pr_reg[CTX_A0],
+			tcp->u_nargs*sizeof(tcp->u_arg[0]));
+	}
+#elif UNIXWARE >= 2
+	/*
+	 * Like SGI, UnixWare doesn't set pr_sysarg until system call exit
+	 */
+	if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+		tcp->u_nargs = sysent[tcp->scno].nargs;
+	else
+		tcp->u_nargs = tcp->status.pr_lwp.pr_nsysarg;
+	umoven(tcp, tcp->status.PR_REG[UESP] + 4,
+		tcp->u_nargs*sizeof(tcp->u_arg[0]), (char *) tcp->u_arg);
+#elif defined (HAVE_PR_SYSCALL)
+	if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+		tcp->u_nargs = sysent[tcp->scno].nargs;
+	else
+		tcp->u_nargs = tcp->status.pr_nsysarg;
+	{
+		int i;
+		for (i = 0; i < tcp->u_nargs; i++)
+			tcp->u_arg[i] = tcp->status.pr_sysarg[i];
+	}
+#elif defined (I386)
+	if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1)
+		tcp->u_nargs = sysent[tcp->scno].nargs;
+	else
+		tcp->u_nargs = 5;
+	umoven(tcp, tcp->status.PR_REG[UESP] + 4,
+		tcp->u_nargs*sizeof(tcp->u_arg[0]), (char *) tcp->u_arg);
+#else
+	I DONT KNOW WHAT TO DO
+#endif /* !HAVE_PR_SYSCALL */
+#endif /* SVR4 */
+#ifdef FREEBSD
+	if (tcp->scno >= 0 && tcp->scno < nsyscalls &&
+	    sysent[tcp->scno].nargs > tcp->status.val)
+		tcp->u_nargs = sysent[tcp->scno].nargs;
+	else 
+	  	tcp->u_nargs = tcp->status.val;
+	if (tcp->u_nargs < 0)
+		tcp->u_nargs = 0;
+	if (tcp->u_nargs > MAX_ARGS)
+		tcp->u_nargs = MAX_ARGS;
+	switch(regs.r_eax) {
+	case SYS___syscall:
+		pread(tcp->pfd, &tcp->u_arg, tcp->u_nargs * sizeof(unsigned long),
+		      regs.r_esp + sizeof(int) + sizeof(quad_t));
+	  break;
+        case SYS_syscall:
+		pread(tcp->pfd, &tcp->u_arg, tcp->u_nargs * sizeof(unsigned long),
+		      regs.r_esp + 2 * sizeof(int));
+	  break;
+        default:
+		pread(tcp->pfd, &tcp->u_arg, tcp->u_nargs * sizeof(unsigned long),
+		      regs.r_esp + sizeof(int));
+	  break;
+	}
+#endif /* FREEBSD */
+	return 1;
+}
+
+int
+trace_syscall(tcp)
+struct tcb *tcp;
+{
+	int sys_res;
+	struct timeval tv;
+	int res;
+
+	/* Measure the exit time as early as possible to avoid errors. */
+	if (dtime && (tcp->flags & TCB_INSYSCALL))
+		gettimeofday(&tv, NULL);
+
+	res = get_scno(tcp);
+	if (res != 1)
+		return res;
+
+	res = syscall_fixup(tcp);
+	if (res != 1)
+		return res;
+
+	if (tcp->flags & TCB_INSYSCALL) {
+		long u_error;
+		res = get_error(tcp);
+		if (res != 1)
+			return res;
+		u_error = tcp->u_error;
+
 
 		internal_syscall(tcp);
-		if (!(qual_flags[tcp->scno] & QUAL_TRACE)) {
+		if (tcp->scno >= 0 && tcp->scno < nsyscalls &&
+		    !(qual_flags[tcp->scno] & QUAL_TRACE)) {
 			tcp->flags &= ~TCB_INSYSCALL;
 			return 0;
 		}
@@ -978,16 +1424,16 @@ struct tcb *tcp;
 		if (tcp->flags & TCB_REPRINT) {
 			printleader(tcp);
 			tprintf("<... ");
-			if (tcp->scno >= nsyscalls)
+			if (tcp->scno >= nsyscalls || tcp->scno < 0)
 				tprintf("syscall_%lu", tcp->scno);
 			else
 				tprintf("%s", sysent[tcp->scno].sys_name);
 			tprintf(" resumed> ");
 		}
 
-		if (cflag) {
+		if (cflag && tcp->scno < nsyscalls && tcp->scno >= 0) {
 			call_count[tcp->scno]++;
-			if (u_error)
+			if (tcp->u_error)
 				error_count[tcp->scno]++;
 			tv_sub(&tv, &tv, &tcp->etime);
 #ifdef LINUX
@@ -1013,7 +1459,7 @@ struct tcb *tcp;
 			return 0;
 		}
 
-		if (tcp->scno >= nsyscalls
+		if (tcp->scno >= nsyscalls || tcp->scno < 0
 		    || (qual_flags[tcp->scno] & QUAL_RAW))
 			sys_res = printargs(tcp);
 		else
@@ -1021,15 +1467,16 @@ struct tcb *tcp;
 		u_error = tcp->u_error;
 		tprintf(") ");
 		tabto(acolumn);
-		if (qual_flags[tcp->scno] & QUAL_RAW) {
+		if (tcp->scno >= nsyscalls || tcp->scno < 0 ||
+		    qual_flags[tcp->scno] & QUAL_RAW) {
 			if (u_error)
 				tprintf("= -1 (errno %ld)", u_error);
 			else
 				tprintf("= %#lx", tcp->u_rval);
 		}
 		else if (!(sys_res & RVAL_NONE) && u_error) {
-#ifdef LINUX
 			switch (u_error) {
+#ifdef LINUX
 			case ERESTARTSYS:
 				tprintf("= ? ERESTARTSYS (To be restarted)");
 				break;
@@ -1039,8 +1486,8 @@ struct tcb *tcp;
 			case ERESTARTNOHAND:
 				tprintf("= ? ERESTARTNOHAND (To be restarted)");
 				break;
-			default:
 #endif /* LINUX */
+			default:
 				tprintf("= -1 ");
 				if (u_error < nerrnos && u_error < sys_nerr)
 					tprintf("%s (%s)", errnoent[u_error],
@@ -1053,10 +1500,8 @@ struct tcb *tcp;
 						sys_errlist[u_error]);
 				else
 					tprintf("E??? (errno %ld)", u_error);
-#ifdef LINUX
 				break;
 			}
-#endif /* LINUX */
 		}
 		else {
 			if (sys_res & RVAL_NONE)
@@ -1075,6 +1520,20 @@ struct tcb *tcp;
 				case RVAL_DECIMAL:
 					tprintf("= %ld", tcp->u_rval);
 					break;
+#ifdef HAVE_LONG_LONG
+				case RVAL_LHEX:
+					tprintf("= %#llx", tcp->u_lrval);
+					break;
+				case RVAL_LOCTAL:
+					tprintf("= %#llo", tcp->u_lrval);
+					break;
+				case RVAL_LUDECIMAL:
+					tprintf("= %llu", tcp->u_lrval);
+					break;
+				case RVAL_LDECIMAL:
+					tprintf("= %lld", tcp->u_lrval);
+					break;
+#endif
 				default:
 					fprintf(stderr,
 						"invalid rval format\n");
@@ -1099,146 +1558,13 @@ struct tcb *tcp;
 	}
 
 	/* Entering system call */
-	tcp->scno = scno;
-#ifdef LINUX
-#if defined(S390)
-	{
-		int i;
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-		for (i = 0; i < tcp->u_nargs; i++) {
-			if (upeek(pid,i==0 ? PT_ORIGGPR2:PT_GPR2+(i<<2), &tcp->u_arg[i]) < 0)
-				return -1;
-		}
-	}
-#elif defined (ALPHA)
-	{
-		int i;
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-		for (i = 0; i < tcp->u_nargs; i++) {
-			/* WTA: if scno is out-of-bounds this will bomb. Add range-check
-			 * for scno somewhere above here!
-			 */
-			if (upeek(pid, REG_A0+i, &tcp->u_arg[i]) < 0)
-				return -1;
-		}
-	}
-#elif defined (MIPS)
-	{
-	  	long sp;
-	  	int i, nargs;
+	res = syscall_enter(tcp);
+	if (res != 1)
+		return res;
 
-		nargs = tcp->u_nargs = sysent[tcp->scno].nargs;
-		if(nargs > 4) {
-		  	if(upeek(pid, REG_SP, &sp) < 0)
-			  	return -1;
-			for(i = 0; i < 4; i++) {
-			  	if (upeek(pid, REG_A0 + i, &tcp->u_arg[i])<0)
-				  	return -1;
-			}
-			umoven(tcp, sp+16, (nargs-4) * sizeof(tcp->u_arg[0]),
-			       (char *)(tcp->u_arg + 4));
-		} else {
-		  	for(i = 0; i < nargs; i++) {
-			  	if (upeek(pid, REG_A0 + i, &tcp->u_arg[i]) < 0)
-				  	return -1;
-			}
-		}
-	}
-#elif defined (POWERPC)
-	{
-		int i;
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-		for (i = 0; i < tcp->u_nargs; i++) {
-			if (upeek(pid, (i==0) ? (4*PT_ORIG_R3) : ((i+PT_R3)*4), &tcp->u_arg[i]) < 0)
-				return -1;
-		}
-	}
-#elif defined (SPARC)
-	{
-		int i;
-	         
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-		for (i = 0; i < tcp->u_nargs; i++)
-			tcp->u_arg[i] = *((&regs.r_o0) + i);
-	}
-#else 
-	{
-		int i;
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-		for (i = 0; i < tcp->u_nargs; i++) {
-			if (upeek(pid, i*4, &tcp->u_arg[i]) < 0)
-				return -1;
-		}
-	}
-#endif 
-#endif /* LINUX */
-#ifdef SUNOS4
-	{
-		int i;
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-		for (i = 0; i < tcp->u_nargs; i++) {
-			struct user *u;
-
-			if (upeek(pid, uoff(u_arg[0]) +
-			    (i*sizeof(u->u_arg[0])), &tcp->u_arg[i]) < 0)
-				return -1;
-		}
-	}
-#endif /* SUNOS4 */
-#ifdef SVR4
-#ifdef MIPS
-	/*
-	 * SGI is broken: even though it has pr_sysarg, it doesn't
-	 * set them on system call entry.  Get a clue.
-	 */
-	if (sysent[tcp->scno].nargs != -1)
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-	else
-		tcp->u_nargs = tcp->status.pr_nsysarg;
-	if (tcp->u_nargs > 4) {
-		memcpy(tcp->u_arg, &tcp->status.pr_reg[CTX_A0],
-			4*sizeof(tcp->u_arg[0]));
-		umoven(tcp, tcp->status.pr_reg[CTX_SP] + 16,
-			(tcp->u_nargs - 4)*sizeof(tcp->u_arg[0]), (char *) (tcp->u_arg + 4));
-	}
-	else {
-		memcpy(tcp->u_arg, &tcp->status.pr_reg[CTX_A0],
-			tcp->u_nargs*sizeof(tcp->u_arg[0]));
-	}
-#else /* !MIPS */
-#ifdef HAVE_PR_SYSCALL
-	if (sysent[tcp->scno].nargs != -1)
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-	else
-		tcp->u_nargs = tcp->status.pr_nsysarg;
-	{
-		int i;
-		for (i = 0; i < tcp->u_nargs; i++)
-			tcp->u_arg[i] = tcp->status.pr_sysarg[i];
-	}
-#else /* !HAVE_PR_SYSCALL */
-#ifdef I386
-	if (sysent[tcp->scno].nargs != -1)
-		tcp->u_nargs = sysent[tcp->scno].nargs;
-	else
-#if UNIXWARE >= 2
-		tcp->u_nargs = tcp->status.pr_lwp.pr_nsysarg;
-#else
-		tcp->u_nargs = 5;
-#endif
-	umoven(tcp, tcp->status.PR_REG[UESP] + 4,
-		tcp->u_nargs*sizeof(tcp->u_arg[0]), (char *) tcp->u_arg);
-#endif /* I386 */
-#endif /* !HAVE_PR_SYSCALL */
-#endif /* !MIPS */
-#endif /* SVR4 */
-#ifdef __arm__
-	switch (tcp->scno + __NR_SYSCALL_BASE) {
-#else
-	switch (tcp->scno) {
-#endif
+	switch (tcp->scno + NR_SYSCALL_BASE) {
 #ifdef LINUX
-#if !defined (ALPHA) && !defined(SPARC) && !defined(MIPS)
+#if !defined (ALPHA) && !defined(IA64) && !defined(SPARC) && !defined(MIPS) && !defined(HPPA)
 	case SYS_socketcall:
 		decode_subcall(tcp, SYS_socket_subcall,
 			SYS_socket_nsubcalls, deref_style);
@@ -1247,7 +1573,7 @@ struct tcb *tcp;
 		decode_subcall(tcp, SYS_ipc_subcall,
 			SYS_ipc_nsubcalls, shift_style);
 		break;
-#endif /* !ALPHA && !SPARC */
+#endif /* !ALPHA && !IA64 && !MIPS && !SPARC */
 #ifdef SPARC
 	case SYS_socketcall:
 		sparc_socket_decode (tcp);
@@ -1305,7 +1631,20 @@ struct tcb *tcp;
 			SYS_door_nsubcalls, door_style);
 		break;
 #endif /* SYS_door_subcall */
+#ifdef SYS_kaio_subcall
+	case SYS_kaio:
+		decode_subcall(tcp, SYS_kaio_subcall,
+			SYS_kaio_nsubcalls, shift_style);
+		break;
+#endif	
 #endif /* SVR4 */
+#ifdef FREEBSD
+	case SYS_msgsys:
+	case SYS_shmsys:
+	case SYS_semsys:
+		decode_subcall(tcp, 0, 0, table_style);
+		break;
+#endif		
 #ifdef SUNOS4
 	case SYS_semsys:
 		decode_subcall(tcp, SYS_semsys_subcall,
@@ -1323,7 +1662,7 @@ struct tcb *tcp;
 	}
 
 	internal_syscall(tcp);
-	if (!(qual_flags[tcp->scno] & QUAL_TRACE)) {
+	if (tcp->scno >=0 && tcp->scno < nsyscalls && !(qual_flags[tcp->scno] & QUAL_TRACE)) {
 		tcp->flags |= TCB_INSYSCALL;
 		return 0;
 	}
@@ -1337,11 +1676,11 @@ struct tcb *tcp;
 	printleader(tcp);
 	tcp->flags &= ~TCB_REPRINT;
 	tcp_last = tcp;
-	if (tcp->scno >= nsyscalls)
+	if (tcp->scno >= nsyscalls || tcp->scno < 0)
 		tprintf("syscall_%lu(", tcp->scno);
 	else
 		tprintf("%s(", sysent[tcp->scno].sys_name);
-	if (tcp->scno >= nsyscalls ||
+	if (tcp->scno >= nsyscalls || tcp->scno < 0 || 
 	    ((qual_flags[tcp->scno] & QUAL_RAW) && tcp->scno != SYS_exit))
 		sys_res = printargs(tcp);
 	else
@@ -1399,7 +1738,11 @@ struct tcb *tcp;
 	val = tcp->status.PR_REG[CTX_V1];
 #endif /* MIPS */
 #endif /* SVR4 */
-
+#ifdef FREEBSD
+	struct reg regs;
+	pread(tcp->pfd_reg, &regs, sizeof(regs), 0);
+	val = regs.r_edx;
+#endif	
 	return val;
 }
 
