@@ -27,7 +27,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: system.c,v 1.13 1999/12/23 14:20:15 wichert Exp $
+ *	$Id: system.c,v 1.19 2001/02/19 13:35:53 wichert Exp $
  */
 
 #include "defs.h"
@@ -68,9 +68,7 @@ struct pt_regs;
 #include <asm/cachectl.h>
 #endif
 
-#ifdef LINUX
 #include <linux/sysctl.h>
-#endif
 
 static struct xlat mount_flags[] = {
 	{ MS_RDONLY,	"MS_RDONLY"	},
@@ -158,6 +156,50 @@ struct tcb *tcp;
 	return 0;
 }
 
+#include <linux/reboot.h>
+static struct xlat bootflags1[] = {
+	{ LINUX_REBOOT_MAGIC1,	"LINUX_REBOOT_MAGIC1"	},
+	{ 0,			NULL			},
+};
+
+static struct xlat bootflags2[] = {
+	{ LINUX_REBOOT_MAGIC2,	"LINUX_REBOOT_MAGIC2"	},
+	{ LINUX_REBOOT_MAGIC2A,	"LINUX_REBOOT_MAGIC2A"	},
+	{ LINUX_REBOOT_MAGIC2B,	"LINUX_REBOOT_MAGIC2B"	},
+	{ 0,			NULL			},
+};
+
+static struct xlat bootflags3[] = {
+	{ LINUX_REBOOT_CMD_CAD_OFF,	"LINUX_REBOOT_CMD_CAD_OFF"	},
+	{ LINUX_REBOOT_CMD_RESTART,	"LINUX_REBOOT_CMD_RESTART"	},
+	{ LINUX_REBOOT_CMD_HALT,	"LINUX_REBOOT_CMD_HALT"		},
+	{ LINUX_REBOOT_CMD_CAD_ON,	"LINUX_REBOOT_CMD_CAD_ON"	},
+	{ LINUX_REBOOT_CMD_POWER_OFF,	"LINUX_REBOOT_CMD_POWER_OFF"	},
+	{ LINUX_REBOOT_CMD_RESTART2,	"LINUX_REBOOT_CMD_RESTART2"	},
+	{ 0,				NULL				},
+};
+
+int
+sys_reboot(tcp)
+struct tcb *tcp;
+{
+	if (entering(tcp)) {
+		if (!printflags(bootflags1, tcp->u_arg[0]))
+			tprintf("LINUX_REBOOT_MAGIC???");
+		tprintf(", ");
+		if (!printflags(bootflags2, tcp->u_arg[1]))
+			tprintf("LINUX_REBOOT_MAGIC???");
+		tprintf(", ");
+		if (!printflags(bootflags3, tcp->u_arg[2]))
+			tprintf("LINUX_REBOOT_CMD_???");
+		if (tcp->u_arg[2] == LINUX_REBOOT_CMD_RESTART2) {
+			tprintf(", ");
+			printstr(tcp, tcp->u_arg[3], -1);
+		}
+	}
+	return 0;
+}
+
 #ifdef M68K
 static struct xlat cacheflush_scope[] = {
 #ifdef FLUSH_SCOPE_LINE
@@ -501,6 +543,19 @@ static struct xlat sysconflimits[] = {
 	{ 0,		NULL		},
 };
 
+int
+sys_sysconf(tcp)
+struct tcb *tcp;
+{
+	if (entering(tcp)) {
+		printxval(sysconflimits, tcp->u_arg[0], "_SC_???");
+	}
+	return 0;
+}
+
+#endif /* SUNOS4 */
+
+#if defined(SUNOS4) || defined(FREEBSD)
 static struct xlat pathconflimits[] = {
 #ifdef	_PC_LINK_MAX
 	{ _PC_LINK_MAX,	"_PC_LINK_MAX"	},	/* max links to file/dir */
@@ -532,15 +587,6 @@ static struct xlat pathconflimits[] = {
 	{ 0,		NULL		},
 };
 
-int
-sys_sysconf(tcp)
-struct tcb *tcp;
-{
-	if (entering(tcp)) {
-		printxval(sysconflimits, tcp->u_arg[0], "_SC_???");
-	}
-	return 0;
-}
 
 int
 sys_pathconf(tcp)
@@ -549,7 +595,7 @@ struct tcb *tcp;
 	if (entering(tcp)) {
 		printstr(tcp, tcp->u_arg[0], -1);
 		tprintf(", ");
-		printxval(pathconflimits, tcp->u_arg[1], "_SC_???");
+		printxval(pathconflimits, tcp->u_arg[1], "_PC_???");
 	}
 	return 0;
 }
@@ -560,12 +606,12 @@ struct tcb *tcp;
 {
 	if (entering(tcp)) {
 		tprintf("%lu, ", tcp->u_arg[0]);
-		printxval(pathconflimits, tcp->u_arg[1], "_SC_???");
+		printxval(pathconflimits, tcp->u_arg[1], "_PC_???");
 	}
 	return 0;
 }
 
-#endif /* SUNOS4 */
+#endif /* SUNOS4 || FREEBSD */
 
 #ifdef SVR4
 
@@ -1364,19 +1410,87 @@ struct tcb *tcp;
 #endif /* SVR4 */
 
 #ifdef SYS_capget
+
+static struct xlat capabilities[] = {
+	{ 1<<CAP_CHOWN,		"CAP_CHOWN"	},
+	{ 1<<CAP_DAC_OVERRIDE,	"CAP_DAC_OVERRIDE"},
+	{ 1<<CAP_DAC_READ_SEARCH,"CAP_DAC_READ_SEARCH"},
+	{ 1<<CAP_FOWNER,	"CAP_FOWNER"	},
+	{ 1<<CAP_FSETID,	"CAP_FSETID"	},
+	{ 1<<CAP_KILL,		"CAP_KILL"	},
+	{ 1<<CAP_SETGID,	"CAP_SETGID"	},
+	{ 1<<CAP_SETUID,	"CAP_SETUID"	},
+	{ 1<<CAP_SETPCAP,	"CAP_SETPCAP"	},
+	{ 1<<CAP_LINUX_IMMUTABLE,"CAP_LINUX_IMMUTABLE"},
+	{ 1<<CAP_NET_BIND_SERVICE,"CAP_NET_BIND_SERVICE"},
+	{ 1<<CAP_NET_BROADCAST,	"CAP_NET_BROADCAST"},
+	{ 1<<CAP_NET_ADMIN,	"CAP_NET_ADMIN"	},
+	{ 1<<CAP_NET_RAW,	"CAP_NET_RAW"	},
+	{ 1<<CAP_IPC_LOCK,	"CAP_IPC_LOCK"	},
+	{ 1<<CAP_IPC_OWNER,	"CAP_IPC_OWNER"	},
+	{ 1<<CAP_SYS_MODULE,	"CAP_SYS_MODULE"},
+	{ 1<<CAP_SYS_RAWIO,	"CAP_SYS_RAWIO"	},
+	{ 1<<CAP_SYS_CHROOT,	"CAP_SYS_CHROOT"},
+	{ 1<<CAP_SYS_PTRACE,	"CAP_SYS_PTRACE"},
+	{ 1<<CAP_SYS_PACCT,	"CAP_SYS_PACCT"	},
+	{ 1<<CAP_SYS_ADMIN,	"CAP_SYS_ADMIN"	},
+	{ 1<<CAP_SYS_BOOT,	"CAP_SYS_BOOT"	},
+	{ 1<<CAP_SYS_NICE,	"CAP_SYS_NICE"	},
+	{ 1<<CAP_SYS_RESOURCE,	"CAP_SYS_RESOURCE"},
+	{ 1<<CAP_SYS_TIME,	"CAP_SYS_TIME"	},
+	{ 1<<CAP_SYS_TTY_CONFIG,"CAP_SYS_TTY_CONFIG"},
+	{ 0,                    NULL            },
+};
+
+
 int
 sys_capget(tcp)
 struct tcb *tcp;
 {
-	cap_user_header_t       arg0;
-	cap_user_data_t         arg1;
+	static cap_user_header_t       arg0 = NULL;
+	static cap_user_data_t         arg1 = NULL;
 
 	if(!entering(tcp)) {
-		arg0 = (cap_user_header_t)tcp->u_arg[0];
-		arg1 = (cap_user_data_t)tcp->u_arg[1];
-		tprintf("{%lx, %d}, ", (unsigned long)arg0->version, arg0->pid);
-		tprintf("{%lx, %lx, %lx}", (unsigned long)arg1->effective,
-			(unsigned long)arg1->permitted, (unsigned long)arg1->inheritable);
+		if (!arg0) {
+			if ((arg0 = malloc(sizeof(*arg0))) == NULL) {
+				fprintf(stderr, "sys_capget: no memory\n");
+				tprintf("%#lx, %#lx", tcp->u_arg[0], tcp->u_arg[1]);
+				return -1;
+			}
+		}
+		if (!arg1) {
+			if ((arg1 = malloc(sizeof(*arg1))) == NULL) {
+				fprintf(stderr, "sys_capget: no memory\n");
+				tprintf("%#lx, %#lx", tcp->u_arg[0], tcp->u_arg[1]);
+				return -1;
+			}
+		}
+
+		if (!tcp->u_arg[0])
+			tprintf("NULL");
+		else if (!verbose(tcp))
+			tprintf("%#lx", tcp->u_arg[0]);
+		else if (umoven(tcp, tcp->u_arg[0], sizeof(*arg0), (char *) arg0) < 0)
+			tprintf("???");
+		else {
+			tprintf("%#x, %d", arg0->version, arg0->pid);
+		}
+		tprintf(", ");
+		if (!tcp->u_arg[1])
+			tprintf("NULL");
+		else if (!verbose(tcp))
+			tprintf("%#lx", tcp->u_arg[1]);
+		else if (umoven(tcp, tcp->u_arg[1], sizeof(*arg1), (char *) arg1) < 0)
+			tprintf("???");
+		else {
+			tprintf("{");
+			printflags(capabilities, arg1->effective);
+			tprintf(", ");
+			printflags(capabilities, arg1->permitted);
+			tprintf(", ");
+			printflags(capabilities, arg1->inheritable);
+			tprintf("}");
+		} 
 	}
 	return 0;
 }
@@ -1385,15 +1499,50 @@ int
 sys_capset(tcp)
 struct tcb *tcp;
 {
-	cap_user_header_t       arg0;
-	cap_user_data_t         arg1;
+	static cap_user_header_t       arg0 = NULL;
+	static cap_user_data_t         arg1 = NULL;
 
-   	if(entering(tcp)) {
-		arg0 = (cap_user_header_t)tcp->u_arg[0];
-		arg1 = (cap_user_data_t)tcp->u_arg[1];
-		tprintf("{%lx, %d}, ", (unsigned long)arg0->version, arg0->pid);
-		tprintf("{%lx, %lx, %lx}", (unsigned long)arg1->effective,
-			(unsigned long)arg1->permitted, (unsigned long)arg1->inheritable);
+	if(entering(tcp)) {
+		if (!arg0) {
+			if ((arg0 = malloc(sizeof(*arg0))) == NULL) {
+				fprintf(stderr, "sys_capset: no memory\n");
+				tprintf("%#lx, %#lx", tcp->u_arg[0], tcp->u_arg[1]);
+				return -1;
+			}
+		}
+		if (!arg1) {
+			if ((arg1 = malloc(sizeof(*arg1))) == NULL) {
+				fprintf(stderr, "sys_capset: no memory\n");
+				tprintf("%#lx, %#lx", tcp->u_arg[0], tcp->u_arg[1]);
+				return -1;
+			}
+		}
+
+		if (!tcp->u_arg[0])
+			tprintf("NULL");
+		else if (!verbose(tcp))
+			tprintf("%#lx", tcp->u_arg[0]);
+		else if (umoven(tcp, tcp->u_arg[0], sizeof(*arg0), (char *) arg0) < 0)
+			tprintf("???");
+		else {
+			tprintf("%#x, %d", arg0->version, arg0->pid);
+		}
+		tprintf(", ");
+		if (!tcp->u_arg[1])
+			tprintf("NULL");
+		else if (!verbose(tcp))
+			tprintf("%#lx", tcp->u_arg[1]);
+		else if (umoven(tcp, tcp->u_arg[1], sizeof(*arg1), (char *) arg1) < 0)
+			tprintf("???");
+		else {
+			tprintf("{");
+			printflags(capabilities, arg1->effective);
+			tprintf(", ");
+			printflags(capabilities, arg1->permitted);
+			tprintf(", ");
+			printflags(capabilities, arg1->inheritable);
+			tprintf("}");
+		} 
 	}
 	return 0;
 }
@@ -1734,7 +1883,7 @@ struct tcb *tcp;
 #endif
 				 )))) {
 			printpath(tcp, (size_t)info.oldval);
-			tprintf(", %d, ", oldlen);
+			tprintf(", %Zu, ", oldlen);
 			if (info.newval == 0)
 				tprintf("NULL");
 			else if (syserror(tcp))
@@ -1743,7 +1892,7 @@ struct tcb *tcp;
 				printpath(tcp, (size_t)info.newval);
 			tprintf(", %Zd", info.newlen);
 		} else {
-			tprintf("%p, %d, %p, %Zd", info.oldval, oldlen,
+			tprintf("%p, %Zd, %p, %Zd", info.oldval, oldlen,
 				info.newval, info.newlen);
 		}
 		tprintf("}");
@@ -1758,3 +1907,136 @@ struct tcb *tcp;
 }
 #endif
 
+#ifdef FREEBSD
+#include <sys/sysctl.h>
+
+int sys___sysctl(tcp)
+struct tcb *tcp;
+{
+	int qoid[CTL_MAXNAME+2];
+	char ctl[1024];
+	size_t len;
+	int i, numeric;
+	
+	if (entering(tcp)) {
+		if (tcp->u_arg[1] < 0 || tcp->u_arg[1] > CTL_MAXNAME ||
+		    (umoven(tcp, tcp->u_arg[0], tcp->u_arg[1] * sizeof(int),
+			    (char *) (qoid + 2)) < 0))
+			tprintf("[...], ");
+		else {
+			/* Use sysctl to ask the name of the current MIB
+			   This uses the undocumented "Staff-functions" used
+			   by the sysctl program. See kern_sysctl.c for
+			   details. */
+			qoid[0] = 0; /* sysctl */
+			qoid[1] = 1; /* name */
+			i = sizeof(ctl);
+			tprintf("[");
+			if (sysctl(qoid, tcp->u_arg[1] + 2, ctl, &i, 0, 0) >= 0) {
+				numeric = !abbrev(tcp);
+				tprintf("%s%s", ctl, numeric ? ", " : "");
+			} else
+				numeric = 1;
+			if (numeric) {
+				for (i = 0; i < tcp->u_arg[1]; i++)
+					tprintf("%s%d", i ? "." : "", qoid[i + 2]);
+			}
+			tprintf("], ");
+			tprintf("%lu, ", tcp->u_arg[1]);
+		}
+	} else {
+		if (!syserror(tcp) && (umove(tcp, tcp->u_arg[3], &len) >= 0)) {
+			printstr(tcp, tcp->u_arg[2], len);
+			tprintf(", [%u], ", len);
+		} else 
+			tprintf("%#lx, %#lx, ", tcp->u_arg[2], tcp->u_arg[3]);
+		printstr(tcp, tcp->u_arg[4], tcp->u_arg[5]);
+		tprintf(", %lu", tcp->u_arg[5]);
+	}
+	return 0;
+}
+#endif
+
+#if UNIXWARE >= 2
+
+#include <sys/ksym.h>
+#include <sys/elf.h>
+
+static struct xlat ksym_flags[] = {
+	{ STT_NOTYPE,	"STT_NOTYPE"	},	
+	{ STT_FUNC,	"STT_FUNC"	},	
+	{ STT_OBJECT,	"STT_OBJECT"	},	
+	{ 0,		NULL		},
+};
+
+int
+sys_getksym(tcp)
+struct tcb *tcp;
+{
+	if (entering (tcp)) {
+		printstr(tcp, tcp->u_arg[0], -1);
+		tprintf(", ");
+	}
+	else {
+		if (syserror(tcp)) {
+			tprintf("%#lx, %#lx",
+				tcp->u_arg[1], tcp->u_arg[2]);
+		}
+		else {
+			int val;
+			printnum (tcp, tcp->u_arg[1], "%#lx");
+			tprintf(", ");
+			if (umove(tcp, tcp->u_arg[2], &val) < 0) {
+				tprintf("%#lx", tcp->u_arg[2]);
+			}
+			else {
+				tprintf("[");
+				printxval (ksym_flags, val, "STT_???");
+				tprintf("]");
+			}
+		}
+	}
+
+	return 0;
+}
+
+#endif /* UNIXWARE > 2 */
+
+#ifdef MIPS
+
+static struct xlat_sysmips[] = {
+	{ SETNAME,		"SETNAME"	},
+	{ FLUSH_CACHE,		"FLUSH_CACHE"	},
+	{ MIPS_FIXADE,		"MIPS_FIXADE"	},
+	{ MIPS_RDNVRAM,		"MIPS_RDNVRAM"	},
+	{ MIPS_ATOMIC_SET,	"MIPS_ATOMIC_SET"	},
+	{ 0, NULL }
+};
+
+int sys_sysmips(tcp)
+struct tcb *tcp;
+{
+	if (entering(tcp)) {
+		printxval(xlat_sysmips, tcp->u_arg[0], "???");
+		tprintf(", ");
+		if (!verbose(tcp)) {
+			tprintf("%d, %d, %d", tcp->u_arg[1], tcp->u_arg[2], tcp->u_arg[3]);
+		} else if (tcp->u_arg[0]==SETNAME) {
+			char nodename[__NEW_UTS_LEN_1];
+			if (umovestr(tcp, tcp->u_arg[1], (__NEW_UTS_LEN_1), nodename) < 0)
+				tprintf(", %#lx", tcp->u_arg[1]);
+			else
+				tprintf(", \"%s\"", nodename);
+		} else if (tcp->u_arg[0]==MIPS_ATOMIC_SET) {
+			tprintf(", %#lx, 0x%x", tcp->u_arg[1], tcp->u_arg[2]);
+		} else if (tcp->u_arg[0]==MIPS_FIXADE) {
+			tprintf(", 0x%x", tcp->u_arg[1]);
+		} else {
+			tprintf("%d, %d, %d", tcp->u_arg[1], tcp->u_arg[2], tcp->u_arg[3]);
+		}
+	}
+
+	return 0;
+}
+
+#endif /* MIPS */
