@@ -30,7 +30,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: util.c,v 1.16 1999/12/24 23:19:31 wakkerma Exp $
+ *	$Id: util.c,v 1.22 2001/03/27 12:17:17 wichert Exp $
  */
 
 #include "defs.h"
@@ -44,17 +44,20 @@
 #include <link.h>
 #endif /* SUNOS4 */
 
-#if defined(linux)
+#if defined(linux) && (__GLIBC__ < 2 || (__GLIBC__ == 2 && __GLIBC_MINOR__ < 1))
 #include <linux/ptrace.h>
 #endif 
 
+#if defined(LINUX) && defined(IA64)
+#include <asm/ptrace_offsets.h>
+#endif
+
 #ifdef HAVE_SYS_REG_H
 #include <sys/reg.h>
 # define PTRACE_PEEKUSR PTRACE_PEEKUSER
-#endif
-
-#ifdef HAVE_SYS_PTRACE_H
-#include <sys/ptrace.h>
+#elif defined(HAVE_LINUX_PTRACE_H)
+#undef PTRACE_SYSCALL
+#include <linux/ptrace.h>
 #endif
 
 #ifdef SUNOS4_KERNEL_ARCH_KLUDGE
@@ -627,7 +630,7 @@ char *laddr;
 #endif /* !oldway */
 #endif /* SUNOS4 */
 
-#ifdef SVR4
+#ifdef USE_PROCFS
 #ifdef HAVE_MP_PROCFS
 	if (pread(tcp->pfd_as, laddr, len, addr) == -1)
 		return -1;
@@ -648,7 +651,7 @@ char *laddr;
 		return -1;
 #endif /* !HAVE_PREAD */
 #endif /* HAVE_MP_PROCFS */
-#endif /* SVR4 */
+#endif /* USE_PROCFS */
 
 	return 0;
 }
@@ -664,9 +667,9 @@ long addr;
 int len;
 char *laddr;
 {
-#ifdef SVR4
+#ifdef USE_PROCFS
 	return umoven(tcp, addr, len, laddr);
-#else /* !SVR4 */
+#else /* !USE_PROCFS */
 	int started = 0;
 	int pid = tcp->pid;
 	int i, n, m;
@@ -716,7 +719,7 @@ char *laddr;
 		addr += sizeof(long), laddr += m, len -= m;
 	}
 	return 0;
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 }
 
 #ifdef LINUX
@@ -818,7 +821,7 @@ char *laddr;
 
 #endif /* SUNOS4 */
 
-#ifndef SVR4
+#ifndef USE_PROCFS
 
 int
 upeek(pid, off, res)
@@ -862,7 +865,7 @@ long *res;
 	return 0;
 }
 
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 
 long
 getpc(tcp)
@@ -874,6 +877,9 @@ struct tcb *tcp;
 #if defined(I386)
 	if (upeek(tcp->pid, 4*EIP, &pc) < 0)
 		return -1;
+#elif defined(IA64)
+	if (upeek(tcp->pid, PT_B0, &pc) < 0)
+		return -1;
 #elif defined(ARM)
 	if (upeek(tcp->pid, 4*15, &pc) < 0)
 		return -1;
@@ -896,6 +902,9 @@ struct tcb *tcp;
 	pc = regs.r_pc;
 #elif defined(S390)
 	if(upeek(tcp->pid,PT_PSWADDR,&pc) < 0)
+#elif defined(HPPA)
+	if(upeek(tcp->pid,PT_IAOQ0,&pc) < 0)
+		return -1;
 #endif
 	return pc;
 #endif /* LINUX */
@@ -919,6 +928,11 @@ struct tcb *tcp;
 	return 0;
 #endif /* SVR4 */
 
+#ifdef FREEBSD
+	struct reg regs;
+	pread(tcp->pfd_reg, &regs, sizeof(regs), 0);
+	return regs.r_eip;
+#endif /* FREEBSD */
 }
 
 void
@@ -935,8 +949,15 @@ struct tcb *tcp;
 		return;
 	}
 	tprintf("[%08lx] ", eip);
-#else /* !I386K */
-#ifdef POWERPC
+#elif defined(IA62)
+	long ip;
+
+	if (upeek(tcp->pid, PT_B0, &ip) < 0) {
+		tprintf("[????????] ");
+		return;
+	}
+	tprintf("[%08lx] ", ip);
+#elif defined(POWERPC)
 	long pc;
 
 	if (upeek(tcp->pid, 4*PT_NIP, &pc) < 0) {
@@ -944,8 +965,7 @@ struct tcb *tcp;
 		return;
 	}
 	tprintf("[%08lx] ", pc);
-#else /* !POWERPC */
-#ifdef M68K
+#elif defined(M68k)
 	long pc;
 
 	if (upeek(tcp->pid, 4*PT_PC, &pc) < 0) {
@@ -953,8 +973,7 @@ struct tcb *tcp;
 		return;
 	}
 	tprintf("[%08lx] ", pc);
-#else /* !M68K */
-#ifdef ALPHA
+#elif defined(ALPHA)
 	long pc;
 
 	if (upeek(tcp->pid, REG_PC, &pc) < 0) {
@@ -962,19 +981,22 @@ struct tcb *tcp;
 		return;
 	}
 	tprintf("[%08lx] ", pc);
-#else /* !ALPHA */
-#ifdef SPARC
+#elif defined(SPARC)
 	struct regs regs;
 	if (ptrace(PTRACE_GETREGS,tcp->pid,(char *)&regs,0) < 0) {
 		tprintf("[????????] ");
 		return;
 	}
 	tprintf("[%08lx] ", regs.r_pc);
-#endif /* SPARC */
-#endif /* ALPHA */
-#endif /* !M68K */
-#endif /* !POWERPC */
-#endif /* !I386 */
+#elif defined(HPPA)
+	long pc;
+
+	if(upeek(tcp->pid,PT_IAOQ0,&pc) < 0) {
+		tprintf ("[????????] ");
+		return;
+	}
+	tprintf("[%08lx] ", pc);
+#endif /* !architecture */
 #endif /* LINUX */
 
 #ifdef SUNOS4
@@ -993,9 +1015,14 @@ struct tcb *tcp;
 	tprintf("[????????] ");
 #endif
 
+#ifdef FREEBSD
+	struct reg regs;
+	pread(tcp->pfd_reg, &regs, sizeof(regs), 0);
+	tprintf("[%08x] ", regs.r_eip);
+#endif /* FREEBSD */
 }
 
-#ifndef SVR4
+#ifndef USE_PROCFS
 
 int
 setbpt(tcp)
@@ -1044,6 +1071,47 @@ struct tcb *tcp;
 	tcp->flags |= TCB_BPTSET;
 
 #else /* !SPARC */
+#ifdef IA64
+	/*
+	 * Our strategy here is to replace the bundle that contained
+	 * the clone() syscall with a bundle of the form:
+	 *
+	 *	{ 1: br 1b; br 1b; br 1b }
+	 *
+	 * This ensures that the newly forked child will loop
+	 * endlessly until we've got a chance to attach to it.
+	 */
+	{
+#		define LOOP0	0x0000100000000017
+#		define LOOP1	0x4000000000200000
+		unsigned long addr, ipsr;
+		pid_t pid;
+
+		pid = tcp->pid;
+		if (upeek(pid, PT_CR_IPSR, &ipsr) < 0)
+			return -1;
+		if (upeek(pid, PT_CR_IIP, &addr) < 0)
+			return -1;
+		tcp->baddr = addr | ((ipsr >> 41) & 0x3);	/* store "ri" in low two bits */
+
+		errno = 0;
+		tcp->inst[0] = ptrace(PTRACE_PEEKTEXT, pid, (char *) addr + 0, 0);
+		tcp->inst[1] = ptrace(PTRACE_PEEKTEXT, pid, (char *) addr + 8, 0);
+		if (errno) {
+			perror("setbpt: ptrace(PTRACE_PEEKTEXT, ...)");
+			return -1;
+		}
+
+		errno = 0;
+		ptrace(PTRACE_POKETEXT, pid, (char *) addr + 0, LOOP0);
+		ptrace(PTRACE_POKETEXT, pid, (char *) addr + 8, LOOP1);
+		if (errno) {
+			perror("setbpt: ptrace(PTRACE_POKETEXT, ...)");
+			return -1;
+		}
+		tcp->flags |= TCB_BPTSET;
+	}
+#else /* !IA64 */
 
 #if defined (I386)
 #define LOOP	0x0000feeb
@@ -1059,6 +1127,8 @@ struct tcb *tcp;
 #define LOOP	0x1000ffff
 #elif defined(S390)
 #define LOOP	0xa7f40000	/* BRC 15,0 */
+#elif defined(HPPA)
+#define LOOP	0xe81f1ff7	/* b,l,n <loc>,r0 */
 #else
 #error unknown architecture
 #endif
@@ -1085,6 +1155,10 @@ struct tcb *tcp;
 #elif defined(S390)
 	if (upeek(tcp->pid,PT_PSWADDR, &tcp->baddr) < 0)
 		return -1;
+#elif defined(HPPA)
+	if (upeek(tcp->pid, PT_IAOQ0, &tcp->baddr) < 0)
+		return -1;
+	tcp->baddr &= ~0x03;
 #else
 #error unknown architecture
 #endif
@@ -1102,6 +1176,7 @@ struct tcb *tcp;
 	}
 	tcp->flags |= TCB_BPTSET;
 
+#endif /* !IA64 */
 #endif /* SPARC */
 #endif /* LINUX */
 
@@ -1163,21 +1238,17 @@ struct tcb *tcp;
 {
 
 #ifdef LINUX
-#ifdef I386
+#if defined(I386)
 	long eip;
-#else /* !I386 */
-#ifdef POWERPC
+#elif defined(POWERPC)
 	long pc;
-#else /* !POWERPC */
-#ifdef M68K
+#elif defined(M68K)
 	long pc;
-#else /* !M68K */
-#ifdef ALPHA
+#elif defined(ALPHA)
 	long pc;
-#endif /* ALPHA */
-#endif /* !M68K */
-#endif /* !POWERPC */
-#endif /* !I386 */
+#elif defined(HPPA)
+	long iaoq;
+#endif /* architecture */
 
 #ifdef SPARC
 	/* Again, we borrow the SunOS breakpoint code. */
@@ -1192,7 +1263,47 @@ struct tcb *tcp;
 		return -1;
 	}
 	tcp->flags &= ~TCB_BPTSET;
-#else /* !SPARC */
+#elif defined(IA64)
+	{
+		unsigned long addr, ipsr;
+		pid_t pid;
+
+		pid = tcp->pid;
+
+		if (upeek(pid, PT_CR_IPSR, &ipsr) < 0)
+			return -1;
+		if (upeek(pid, PT_CR_IIP, &addr) < 0)
+			return -1;
+
+		/* restore original bundle: */
+		errno = 0;
+		ptrace(PTRACE_POKETEXT, pid, (char *) addr + 0, tcp->inst[0]);
+		ptrace(PTRACE_POKETEXT, pid, (char *) addr + 8, tcp->inst[1]);
+		if (errno) {
+			perror("clearbpt: ptrace(PTRACE_POKETEXT, ...)");
+			return -1;
+		}
+
+		/* restore original "ri" in ipsr: */
+		ipsr = (ipsr & ~(0x3ul << 41)) | ((tcp->baddr & 0x3) << 41);
+		errno = 0;
+		ptrace(PTRACE_POKEUSER, pid, (char *) PT_CR_IPSR, ipsr);
+		if (errno) {
+			perror("clrbpt: ptrace(PTRACE_POKEUSER, ...)");
+			return -1;
+		}
+
+		tcp->flags &= ~TCB_BPTSET;
+
+		if (addr != (tcp->baddr & ~0x3)) {
+			/* the breakpoint has not been reached yet.  */
+			if (debug)
+				fprintf(stderr, "NOTE: PC not at bpt (pc %#lx baddr %#lx)\n",
+					addr, tcp->baddr);
+			return 0;
+		}
+	}
+#else /* !IA64  && ! SPARC */
 
 	if (debug)
 		fprintf(stderr, "[%d] clearing bpt\n", tcp->pid);
@@ -1219,8 +1330,7 @@ struct tcb *tcp;
 					eip, tcp->baddr);
 		return 0;
 	}
-#else /* !I386 */
-#ifdef POWERPC
+#elif defined(POWERPC)
 	if (upeek(tcp->pid, 4*PT_NIP, &pc) < 0)
 		return -1;
 	if (pc != tcp->baddr) {
@@ -1230,8 +1340,7 @@ struct tcb *tcp;
 				pc, tcp->baddr);
 		return 0;
 	}
-#else /* !POWERPC */
-#ifdef M68K
+#elif defined(M68K)
 	if (upeek(tcp->pid, 4*PT_PC, &pc) < 0)
 		return -1;
 	if (pc != tcp->baddr) {
@@ -1241,8 +1350,7 @@ struct tcb *tcp;
 				pc, tcp->baddr);
 		return 0;
 	}
-#else /* !M68K */
-#ifdef ALPHA
+#elif defined(ALPHA)
 	if (upeek(tcp->pid, REG_PC, &pc) < 0)
 		return -1;
 	if (pc != tcp->baddr) {
@@ -1252,11 +1360,26 @@ struct tcb *tcp;
 				pc, tcp->baddr);
 		return 0;
 	}
-#endif /* ALPHA */
-#endif /* !M68K */
-#endif /* !POWERPC */
-#endif /* !I386 */
-#endif /* !SPARC */
+#elif defined(HPPA)
+	if (upeek(tcp->pid, PT_IAOQ0, &iaoq) < 0)
+		return -1;
+	iaoq &= ~0x03;
+	if (iaoq != tcp->baddr && iaoq != tcp->baddr + 4) {
+		/* The breakpoint has not been reached yet.  */
+		if (debug)
+			fprintf(stderr, "NOTE: PC not at bpt (iaoq %#lx baddr %#lx)\n",
+				iaoq, tcp->baddr);
+		return 0;
+	}
+	iaoq = tcp->baddr | 3;
+	/* We should be pointing at a 'ldi -1000,r1' in glibc, so it is
+	 * safe to set both IAOQ0 and IAOQ1 to that so the PSW N bit
+	 * has no significant effect.
+	 */
+	ptrace(PTRACE_POKEUSER, tcp->pid, (void *)PT_IAOQ0, iaoq);
+	ptrace(PTRACE_POKEUSER, tcp->pid, (void *)PT_IAOQ1, iaoq);
+#endif /* arch */
+#endif /* !SPARC && !IA64 */
 #endif /* LINUX */
 
 #ifdef SUNOS4
@@ -1312,7 +1435,7 @@ struct tcb *tcp;
 	return 0;
 }
 
-#endif /* !SVR4 */
+#endif /* !USE_PROCFS */
 
 #ifdef SUNOS4
 
